"""qrem.common.convert module contains helper functions, that allow to convert between various formats used to describe circuit labels, 
bitstrings, quantum registers etc.
"""
import numpy as np
from typing import List, Optional, Iterable
from qrem.common import utils

from qrem.common.constants import CIRCUIT_DATA_TYPE as c_type



# =======================================================================================
# BITSTRINGS CONVERSIONS
# =======================================================================================
def bitstring_to_integer(bitstring: str) -> int:
    return int(bitstring, 2)

def integer_to_bitstring(integer: int,
                         number_of_bits: int) -> str:
    """Return binary representation of an integer in a string form

    Parameters
    ----------
    integer: int
        input integer which bytes will be interpreted as bitstring
    number_of_bits: int
        Number of bits.It  can be greater than minimal needed to represent integer (but if the input number of bits
        is smaller then the minimal one, the latter one will be executed )
    """

    return "{0:b}".format(integer).zfill(number_of_bits)

#ndarrays
def bitstring_to_ndarray(bitstring: str, reverse = False) -> np.ndarray:
    """Return numpy array  representation of a bitstring

    Parameters
    ----------
    bitstring: string
        input bitstring
    reverse: bool
        The output array can be reversed in order. Default is False
    """
    
    arr = np.frombuffer(bytes(bitstring,"utf-8"), dtype=np.uint8) - ord('0')
    if reverse:
        return np.flip(arr)
    return arr

def ndarray_to_bitstring(bitarray: np.ndarray, reverse = False) -> str:
    if reverse:
        return np.array2string(np.flip(bitarray),separator="")[1:-1]
    return np.array2string(bitarray,separator="")[1:-1]

#dit-strings
def integer_to_ditstring(integer, base, number_of_dits):
    """Return nnary representation of an integer in a string form
    based on
    https://stackoverflow.com/questions/2267362/how-to-convert-an-integer-to-a-string-in-any-base

    Parameters
    ----------
    integer: int
        input integer value
    base: int
        base for the ditstring
    number_of_dits: int
        Number of bits.It  can be greater than minimal needed to represent integer (but if the input number of bits
        is smaller then the minimal one, the latter one will be executed )
    """

    if integer == 0:
        return [0 for _ in range(number_of_dits)]
    digits = []
    while integer:
        digits.append(int(integer % base))
        integer //= base

    # digits = sorted(digits,key = lambda x: x[0])
    if len(digits) < number_of_dits:
        for _ in range(number_of_dits - len(digits)):
            digits.append(0)
    return digits[::-1]

def negate_bitstring(bitstring):
    '''This function creates a negation of a bitstring (assuming it is in string format)
    Parameters
    ----------
    bitstring: int
        input bitstring
    '''
    return ''.join('1' if x == '0' else '0' for x in bitstring)
    # potentially better version? Is it really useful?
    # def negate_bitstring(bit):
    #     if isinstance(bit,int) or isinstance(bit,float) or isinstance(bit,complex):
    #         if bit==0:
    #             return 1
    #         elif bit==1:
    #             return 0
    #         else:
    #             raise ValueError(f"Wrong bit: {bit}")
    #     elif isinstance(bit,str):
    #         if bit in ['0']:
    #             return '1'
    #         elif bit in ['1']:
    #             return '0'
    #         else:
    #             raise ValueError(f"Wrong bit: {bit}")
    #     else:
    #         raise ValueError(f"Wrong datatype of {bit} - '{type(bit)}'")
    pass

def sort_bitstring(string:str, new_order:Iterable) -> str:
    """Sort bits in string according to new_order"""
    sorted_string = utils.ort_things(list(string), new_order)
    sorted_string = ''.join([str(s) for s in sorted_string])
    return sorted_string

def sort_bitarray(bitarray: np.ndarray, new_order:Iterable) -> np.ndarray:
    """Sort bits in string according to new_order"""
    sorted_array = utils.sort_things(bitarray, new_order)
    array =np.ndarray(sorted_array,dtype=c_type)
    return array


# =======================================================================================
# REGISTERS
# =======================================================================================
#MOVE_TO >> core.utils core.numerics
def get_ditstrings_register(base, number_of_dits):
    """ Return a list of natural numbers fitting in number_of_dits with base dit-representation """
    return [integer_to_ditstring(j, base, number_of_dits=number_of_dits) for j in
            list(range(base ** number_of_dits))]

if __name__ == "__main__":
    a = bitstring_to_ndarray("100101",reverse =True)
    print(a)
    print(ndarray_to_bitstring(a,reverse =True))



    
# =======================================================================================
# QBIT INDICIES
# =======================================================================================
#MOVE_TO >> core.utils
def get_qubit_indices_from_keystring(qubits_string: str,
                                  with_q: Optional[bool] = False) -> List[int]:
    """Return list of qubit indices from the string of the form "q0q1q22q31"
    :param qubits_string (string): string which has the form of "q" followed by qubit index
    :param (optional) with_q (Boolean): specify whether returned indices
                                        should be in form of string with letter

    :return: list of qubit indices:

    depending on value of parameter "with_q" the mapping will be one of the following:

    if with_q:
        'q1q5q13' -> ['q1','q5','q13']
    else:
        'q1q5q13' -> [1,5,13]
    """

    numbers = re.findall(r'\d+', qubits_string)

    if with_q:
        qubits = ['q' + s for s in numbers]
    else:
        qubits = [int(s) for s in numbers]

    return qubits


"""Run IBM Experiment pipeline

This module allows to run a characterization/mitigation/benchmark experiment 
on a chosen IBM quantum machine. It uses a config file, which example can be found in 
<configs/ default_ibm.ini> file. Read the config to understand avialable parameters.


 documentation as specified by the `NumPy
Documentation HOWTO`_. Docstrings may extend over multiple lines. Sections
are created with a section header followed by an underline of equal length.

Example
-------
You can run experiment by (i) providing a given config via -C argument, for example

    $ python run_ibm_experiment.py -C \\path\\to\\experiment\\config.ini

    
Notes
-----
    @authors: 
    @contact:
"""

import os, sys
import qrem
from pathlib import Path
import orjson

from qrem.common.io import date_time_formatted
from qrem.common.printer import qprint, errprint, warprint
from qrem.common.experiment import tomography as tomography
from qrem.types import CircuitCollection
from qrem.common.constants import CIRCUIT_DATA_TYPE as c_type
from qrem.common.providers import ibm   #PP: change to qrem.ibm

# ----------------------------------------------------------------
# this are the parameters of running qrem in future / maybe move to config[GENERAL]?:
# you will be able to run it asa command
# ----------------------------------------------------------------
CONFIG_PATH = ["--config_file", "C:\\CFT Chmura\\Theory of Quantum Computation\\QREM_Data\\ibm\\experiment_results\\first_experiment\\first_experiment_config.ini"]

#TODO (PP) write version of runner that works with existing circuit collection file

def run(cmd_args=CONFIG_PATH, verbose_log = True): 
    # ----------------------------------------------------------------
    #[1] We load an already defined default config
    # ----------------------------------------------------------------
    config = qrem.load_config(cmd_args = cmd_args, verbose_log = verbose_log)

    EXPERIMENT_NAME = config.experiment_name

    EXPERIMENT_FOLDER_PATH = Path(config.experiment_path)
    if not EXPERIMENT_FOLDER_PATH.is_dir():
        EXPERIMENT_FOLDER_PATH.mkdir( parents=True, exist_ok=True )

    BACKUP_CIRCUITS = config.backup_circuits
    BACKUP_JOB_IDs = config.backup_job_ids
    BACKUP_CIRCUITS_METADATA = config.backup_circuits_metadata
    CONNECTION_METHOD = config.ibm_connection_method
    JOB_TAGS = list(config.job_tags)

    # ----------------------------------------------------------------
    #[2] Get info from provider about valid qubits
    # ----------------------------------------------------------------
    backend, service, provider = ibm.connect(name = config.device_name, method = CONNECTION_METHOD, verbose_log = config.verbose_log)
    valid_qubit_properties = ibm.get_valid_qubits_properties(backend, config.gate_threshold)

    METADATA = {}
    METADATA["date"] = date_time_formatted()
    if BACKUP_CIRCUITS_METADATA:
        METADATA["valid_qubit_properties"] = valid_qubit_properties

    number_of_qubits = valid_qubit_properties["number_of_good_qubits"]
    good_qubits_indices = valid_qubit_properties["good_qubits_indices"]

    # ----------------------------------------------------------------
    #[3] TODO (PP) check if there are specific qubits subset defined in config,. if yes - check against good_qubits_indices and choose only the subset
    # ----------------------------------------------------------------

    # ----------------------------------------------------------------
    #[4] Generate circuits
    # ----------------------------------------------------------------
    #[4.0] Setup circuit collection object
    qrem_circuit_collection = CircuitCollection(EXPERIMENT_NAME)

    qrem_circuit_collection.load_config(config=config)
    qrem_circuit_collection.qubit_indices = good_qubits_indices
    qrem_circuit_collection.metadata = METADATA

    #[4.1] TODO (PP) finish implementation of generate_circuits, so you know how many shots per circuit you need
    # - TODO (PP) initialize with config object
    qrem_circuit_collection.circuits, _, theoretical_total_circuit_count, theoretical_number_of_shots = tomography.generate_circuits(   number_of_qubits = number_of_qubits,
                                                                            experiment_type = config.experiment_type, 
                                                                            k_locality = config.k_locality,
                                                                            limited_circuit_randomness = config.limited_circuit_randomness,
                                                                            imposed_max_random_circuit_count = config.random_circuits_count,
                                                                            imposed_max_number_of_shots = config.shots_per_circuit)
    
    # - TODO (PP) save final number of shots to collection
    # qrem_circuit_collection.no_shots = ?

    #[5] At this stage, circuits are ready, we can save them to file
    if BACKUP_CIRCUITS:    
        qrem_circuit_collection.export_json(str(EXPERIMENT_FOLDER_PATH.joinpath("input_circuit_collection.json")), overwrite = True)

    #[6] Now we need to use ibm module to prepare circuits in ibm format
    ibm_circuits = ibm.translate_circuits_to_qiskit_format(qrem_circuit_collection)



    #[6] Now we need to run circuits
    #- TODO (PP) initialize with config object?
    job_ids = ibm.execute_circuits( qiskit_circuits= ibm_circuits,
                                    job_tags = JOB_TAGS,
                                    number_of_repetitions = config.shots_per_circuit,
                                    instance = config.provider_instance,
                                    service = service,
                                    backend = backend,
                                    method = CONNECTION_METHOD,
                                    log_level='INFO',
                                    verbose_log=True)

    #[6] Save job ids to a file
    if BACKUP_JOB_IDs: 
        json_job_ids=orjson.dumps(job_ids)
        with open(str(EXPERIMENT_FOLDER_PATH.joinpath("input_circuit_collection.json")), 'wb') as outfile:
            outfile.write(json_job_ids)



if __name__ == "__main__":
    warprint("=============================")
    warprint("START: Executing IBM Experiment pipeline")
    warprint("=============================")
    argv = sys.argv[1:]
    import time
    t0 = time.time()

    if argv != []:
        run(cmd_args=argv)
    else:
        run()

    t1 = time.time()
    qprint("=============================")
    qprint("Elapsed time: {} sec".format(t1 - t0))


# ----------------------------------------------------------------
# how to run:
# ----------------------------------------------------------------
    #python run_ibm_experiment.py -c path/to/config.ini
    #python run_ibm_experiment.py
















#------------------------------------
#------------------------------------
#------------------------------------
#------------------------------------
#qrem.run_characterisation_experiment()
#? qrem.run_mitigation_experiment()
#qrem.download_characterisation_experiment_results()
#qrem.download_mitigation_experiment_results()
#qrem.run_characterisation()
#qrem.run_mitigation()




"""Ncpol2SDPA
=====

Provides
 1. A converter from a polynomial optimization problems of commuting and
    noncommuting variables to a semidefinite programming relaxation.
 2. Helper functions to define physics problems.

"""
__version__ = "1.12.3"
from .faacets_relaxation import FaacetsRelaxation
from .sdp_relaxation import SdpRelaxation
from .steering_hierarchy import SteeringHierarchy
from .moroder_hierarchy import MoroderHierarchy
from .rdm_hierarchy import RdmHierarchy
from .nc_utils import generate_operators, generate_variables, get_monomials, \
                      flatten
from .sdpa_utils import read_sdpa_out
from .physics_utils import bosonic_constraints, fermionic_constraints, \
    pauli_constraints, get_neighbors, get_next_neighbors, correlator, \
    generate_measurements, projective_measurement_constraints, \
    maximum_violation, define_objective_with_I, Probability

__all__ = ['SdpRelaxation',
           'SteeringHierarchy',
           'MoroderHierarchy',
           'FaacetsRelaxation',
           'RdmHierarchy',
           'generate_operators',
           'generate_variables',
           'bosonic_constraints',
           'fermionic_constraints',
           'projective_measurement_constraints',
           'correlator',
           'maximum_violation',
           'generate_measurements',
           'define_objective_with_I',
           'Probability',
           'flatten',
           'get_monomials',
           'get_neighbors',
           'get_next_neighbors',
           'read_sdpa_out',
           'pauli_constraints']



"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com

REFERENCES:
[0] Filip B. Maciejewski, Zoltán Zimborás, Michał Oszmaniec,
"Mitigation of readout noise in near-term quantum devices
by classical post-processing based on detector tomography",
Quantum 4, 257 (2020)

[0.5] Filip B. Maciejewski, Flavio Baccari, Zoltán Zimborás, Michał Oszmaniec,
"Modeling and mitigation of cross-talk effects in readout noise
with applications to the Quantum Approximate Optimization Algorithm",
Quantum 5, 464 (2021).

"""
import copy
from typing import Optional, List, Dict, Union
import time

import numpy as np
from scipy.special import binom as binomial_coefficient
from tqdm import tqdm

from qrem.noise_characterization.tomography_design.overlapping.LabelsBaseDDOT import LabelsBaseDDOT
from qrem.functions_qrem import ancillary_functions as anf

import qrem.common.math as qrem_math
from qrem.common.printer import qprint

class LabelsCreatorDDOT(LabelsBaseDDOT):
    """
    Class for creating symbolic description of Diagonal Detector Overlapping Tomography (DDOT)
    circuits.

    Description of a single circuit is simply a bitstring denoting classical input state.


    """

    def __init__(self,
                 number_of_qubits: int,
                 maximal_circuits_amount: Optional[int] = 1500,
                 subsets_list: Optional[List[List[int]]] = None,
                 subsets_locality: Optional[int] = None,
                 show_progress_bars: Optional[bool] = True
                 ):
        """
        :param number_of_qubits:
        :param subsets_list: list of subsets of qubits for which marginal tomography
                             should be performed
                             NOTE: for tomography which is to be performed NOT on ALL subsets
                             of given locality, this must be provided

        :param subsets_locality: locality of ALL considered subsets
                                 NOTE: this should be provided only if ALL subsets have the same
                                       locality
        """

        if subsets_list is None and subsets_locality is None:
            raise ValueError('Please provide subsets list or desired locality')
        elif subsets_list is None and subsets_locality is not None:
            subsets_list = qrem_math.get_k_local_subsets(number_of_elements=number_of_qubits, 
                                                         subset_size=subsets_locality)

        super().__init__(number_of_qubits=number_of_qubits,
                         subsets=subsets_list,
                         maximal_circuits_amount=maximal_circuits_amount,
                         show_progress_bars=show_progress_bars
                         )

        self._subsets_locality = subsets_locality
        if subsets_locality is not None:
            self._elements_in_perfect_collection = int(
                binomial_coefficient(self._number_of_qubits,
                                     self._subsets_locality) *
                2 ** self._subsets_locality)
        else:
            self._elements_in_perfect_collection = int(sum(
                [2 ** len(subset_now) for subset_now in subsets_list]))

        dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset_list): np.zeros(2 ** len(subset_list),
                                                       dtype=int) for
            subset_list in self._subsets}

        self._dictionary_symbols_counting = dictionary_symbols_counting

        circuits_properties_dictionary = {'number_of_circuits': 0,
                                          'absent_elements_amount': 10 ** 6,
                                          'minimal_amount': 10 ** 6,
                                          'maximal_amount': 10 ** 6,
                                          'median': 10 ** 6,
                                          'mean': 10 ** 6,
                                          'SD': 10 ** 6}

        self._circuits_properties_dictionary = circuits_properties_dictionary

    @property
    def dictionary_symbols_counting(self) -> Dict[str, np.ndarray]:
        return self._dictionary_symbols_counting

    @property
    def circuits_properties_dictionary(self) -> Dict[str, float]:
        return self._circuits_properties_dictionary

    @property
    def circuits_list(self) -> List[List[int]]:
        return self._circuits_list

    @circuits_list.setter
    def circuits_list(self, circuits_list: List[List[int]]):
        self.reset_everything()
        self._circuits_list = circuits_list

    def add_dictionary_subsets_symbols_counting_template(self):
        dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset_list): np.zeros(2 ** len(subset_list),
                                                       dtype=int) for
            subset_list in self._subsets}

        self._dictionary_symbols_counting = dictionary_symbols_counting

    def update_dictionary_subsets_symbols_counting(self,
                                                   circuits_list: Optional[List[int]] = None,
                                                   count_added_subcircuits: Optional[bool] = False):
        if circuits_list is None:
            circuits_list = self._circuits_list

        subsets_range = range(len(self._subsets))
        if self._show_progress_bars:
            subsets_range = tqdm(subsets_range)

        if count_added_subcircuits:
            added_subcircuits_counter = 0
        else:
            added_subcircuits_counter = None

        for subset_index in subsets_range:
            for circuit in circuits_list:

                qubits_key = anf.get_qubits_keystring(self._subsets[subset_index])
                subset_circuit_identifier = int(
                    ''.join([str(circuit[qubit_index]) for qubit_index in
                             self._subsets[subset_index]]),2)

                if count_added_subcircuits:
                    if self._dictionary_symbols_counting[qubits_key][subset_circuit_identifier] == 0:
                        added_subcircuits_counter += 1

                self._dictionary_symbols_counting[qubits_key][subset_circuit_identifier] += 1

        if count_added_subcircuits:
            return added_subcircuits_counter

    def get_absent_symbols_amount(self):

        t0 = time.time()
        zero_subsets = 0
        for subset in self._subsets:
            subset_counts = self._dictionary_symbols_counting[anf.get_qubits_keystring(subset)]
            zero_subsets += len(subset_counts) - np.count_nonzero(subset_counts)

        self._circuits_properties_dictionary['absent_elements_amount'] = zero_subsets
        qprint('This took:', time.time() - t0)

        return zero_subsets

    def calculate_properties_of_circuits(self) -> None:

        big_list = []
        for subset in self._subsets:
            big_list += list(self._dictionary_symbols_counting[anf.get_qubits_keystring(subset)])

        minimal_amount, maximal_amount = min(big_list), max(big_list)

        big_list = np.array(big_list)

        mean, SD, median = np.mean(big_list), np.std(big_list), np.median(big_list)

        absent_elements_amount = len(big_list) - np.count_nonzero(big_list)

        self._circuits_properties_dictionary['number_of_circuits'] = len(self._circuits_list)
        self._circuits_properties_dictionary['absent_elements_amount'] = absent_elements_amount
        self._circuits_properties_dictionary['minimal_amount'] = minimal_amount
        self._circuits_properties_dictionary['maximal_amount'] = maximal_amount

        self._circuits_properties_dictionary['median'] = median
        self._circuits_properties_dictionary['mean'] = mean
        self._circuits_properties_dictionary['SD'] = SD/mean

    def reset_everything(self):
        circuits_properties_dictionary = {'number_of_circuits': 0,
                                          'absent_elements_amount': 10 ** 6,
                                          'minimal_amount': 10 ** 6,
                                          'maximal_amount': 10 ** 6,
                                          'median': 10 ** 6,
                                          'mean': 10 ** 6,
                                          'SD': 10 ** 6}

        self._circuits_properties_dictionary = circuits_properties_dictionary
        self._circuits_list = []

        dictionary_symbols_counting = {
            anf.get_qubits_keystring(subset_list): np.zeros(2 ** len(subset_list),
                                                       dtype=int) for
            subset_list in self._subsets}

        self._dictionary_symbols_counting = dictionary_symbols_counting

    def _cost_function_circuits_amount(self):
        return len(self._circuits_list)

    def _cost_function_circuits_SD(self):
        return self._circuits_properties_dictionary['SD']

    def _cost_function_minimal_amount_of_circuits(self):
        return -self._circuits_properties_dictionary['minimal_amount']

    def _cost_function_absent_elements(self):
        return self._circuits_properties_dictionary['absent_elements_amount']

    def _cost_function_maximal_spread(self):
        return self._circuits_properties_dictionary['maximal_amount'] - \
               self._circuits_properties_dictionary['minimal_amount']

    def __compute_1q_collection(self):

        circuits_now = [[unique_symbol for _ in range(self._number_of_qubits)] for unique_symbol in [0,1]]
        self.add_circuits(circuits_now)

        self.update_dictionary_subsets_symbols_counting(
            circuits_list=circuits_now,
            count_added_subcircuits=True)




    def _compute_perfect_collection_bruteforce(self,
                                               circuits_in_batch: int,
                                               print_updates: bool
                                               ):

        runs_number = 1
        absent_elements_amount = self._elements_in_perfect_collection

        while absent_elements_amount > 0 and runs_number < self._maximal_circuits_amount:
            if runs_number % 20 == 0 and print_updates:
                qprint('Run number:', runs_number)
                qprint('Number of circuits:', len(self._circuits_list))
                qprint('Absent elements amount:', absent_elements_amount)

            circuits_now = self.get_random_circuits_list(circuits_in_batch)
            self.add_circuits(circuits_now)

            added_elements = self.update_dictionary_subsets_symbols_counting(
                circuits_list=circuits_now,
                count_added_subcircuits=True)
            absent_elements_amount -= added_elements

            runs_number += 1

    def _get_proper_cost_function(self,
                                  optimized_quantity: str):

        if optimized_quantity in ['circuits_amount', 'circuits_number', 'amount', 'circuits']:
            cost_function = self._cost_function_circuits_amount
        elif optimized_quantity in ['std', 'SD', 'standard_deviation']:
            cost_function = self._cost_function_circuits_SD
        elif optimized_quantity in ['minimal_amount']:
            cost_function = self._cost_function_minimal_amount_of_circuits
        elif optimized_quantity in ['spread', 'maximal_spread']:
            cost_function = self._cost_function_maximal_spread
        elif optimized_quantity in ['absent', 'absent_elements']:
            cost_function = self._cost_function_absent_elements
        else:
            raise ValueError('Wrong optimized quantity string: ' + optimized_quantity + '.')

        return cost_function

    def _add_cost_functions(self,
                            dictionary_cost_functions: Dict[str, float]):

        def cost_functions_added():
            returned_quantity = 0
            for function_label, function_weight in dictionary_cost_functions.items():
                returned_quantity += function_weight * self._get_proper_cost_function(function_label)()
            return returned_quantity

        return cost_functions_added

    def _compute_perfect_collection_bruteforce_randomized(self,
                                                          number_of_iterations: int,
                                                          circuits_in_batch: int,
                                                          print_updates: bool,
                                                          optimized_quantity: Union[
                                                              str, Dict[str, float]],
                                                          additional_circuits: Optional[int] = 0
                                                          ):
        """
        This function implements self._compute_perfect_collection_bruteforce
        for number_of_iterations times, then adds additional_circuits number of random circuits,
        computes cost function and chooses the family that minimizes cost function.

        :param number_of_iterations: how many times random perfect family should be generated
        :param circuits_in_batch: see self._compute_perfect_collection_bruteforce
        :param print_updates: whether to print updates during optimization
        :param optimized_quantity: specify what cost function is
        Possible string values:
        1. 'minimal_amount' - maximizes number of least-frequent subset-circuits
        2. 'spread' - minimizes difference between maximal and minimal number of subset-circuits

        3. 'circuits_amount' - minimizes number of circuits
                        (NOTE: it does not make sense to choose this option with additional_circuits>0)
        4. 'SD' - minimizes standard deviation of occurrences of subset-circuits

        It is possible to use combined cost functions_qrem.
        Dictionary must be provided where KEY is label for optimized quantity and VALUE is its weight.

        For example:
        optimized_quantity = {'minimal_amount': 1.0,
                              'spread':0.5}

        will give cost function which returns 1.0 * (-number of least frequent circuits)
                                          + 0.5 * (difference between most and least fequenet circuits)


        :param additional_circuits: number of circuits which are to be added to the PERFECT collection
                                    obtained in optimization loop. Those are "additional" circuits in
                                    a sense that they are not needed for collection to be perfect,
                                    but instead are used to obtain better values of cost function
                                    or just add more experiments reduce statistical noise.
        :return:
        """

        if isinstance(optimized_quantity, str):
            cost_function = self._get_proper_cost_function(optimized_quantity=optimized_quantity)
        elif isinstance(optimized_quantity, dict):
            cost_function = self._add_cost_functions(dictionary_cost_functions=optimized_quantity)

        runs_range = range(number_of_iterations)
        if self._show_progress_bars:
            runs_range = tqdm(runs_range)

        # circuit_families = []
        # best_family = None
        global_cost, best_family = 10 ** 6, None

        for runs_number in runs_range:
            if runs_number % int(np.ceil(number_of_iterations / 20)) == 0 and print_updates:
                qprint('Run number:', runs_number, 'red')
                qprint('Current best value of cost function:', global_cost)

            self.reset_everything()
            self._compute_perfect_collection_bruteforce(circuits_in_batch=circuits_in_batch,
                                                        print_updates=False
                                                        )

            if additional_circuits > 0:
                current_length, maximal_length = len(
                    self._circuits_list), self._maximal_circuits_amount
                if current_length < maximal_length:
                    if additional_circuits > maximal_length - current_length:
                        adding_circuits = maximal_length - current_length
                    else:
                        adding_circuits = copy.deepcopy(additional_circuits)
                    # print(adding_circuits)
                    new_circuits = self.get_random_circuits_list(adding_circuits)

                    self.update_dictionary_subsets_symbols_counting(new_circuits)
                    self.add_circuits(new_circuits)

            self.calculate_properties_of_circuits()

            cost_now = cost_function()

            if cost_now < global_cost:
                best_family = copy.deepcopy(self._circuits_list)
                global_cost = cost_now

        qprint('best family length', len(best_family), 'red')
        self.reset_everything()
        self._circuits_list = best_family
        self.update_dictionary_subsets_symbols_counting()
        self.calculate_properties_of_circuits()







    def compute_perfect_collection(self,
                                   method_name='bruteforce_randomized',
                                   method_kwargs=None):

        """
        Find perfect collection of overlapping circuits.
        "Perfect" means that for each subset of qubits self._subsets_list[i],
         each symbol out of 2^self._subsets_list[i],
         appears in the collection at least once.

        :param method_name:
        possible values:

        1. 'bruteforce' - see self._compute_perfect_collection_bruteforce
        2. 'bruteforce_randomized' - see self._compute_perfect_collection_bruteforce_randomized

        :param method_kwargs: kwargs for chosen method, see corresponding methods' descriptions
        :return:
        """


        if self._subsets_locality ==1:
            self.__compute_1q_collection()

        else:

            if method_name == 'bruteforce':
                if method_kwargs is None:
                    method_kwargs = {'circuits_in_batch': 1,
                                     'print_updates': True}

                self._compute_perfect_collection_bruteforce(**method_kwargs)

            elif method_name == 'bruteforce_randomized':
                if method_kwargs is None:
                    method_kwargs = {'number_of_iterations': 100,
                                     'circuits_in_batch': 1,
                                     'print_updates': True,
                                     'optimized_quantity': 'minimal_amount',
                                     'additional_circuits': 0}

                self._compute_perfect_collection_bruteforce_randomized(**method_kwargs)

            absent_elements = self._circuits_properties_dictionary['absent_elements_amount']
            if absent_elements != 0:
                qprint('________WARNING________:',
                               'The collection is not perfect. '
                               'It is missing %s' % absent_elements
                               + ' elements!\nTry increasing limit on the circuits amount.',
                               'red')





import matplotlib.pyplot as plt

#TODO_PP ask Asia which functionalities are covered by her module
# ORGANIZE: these functions can all be deleted; they are either copied to functions_benchmarks or not used anywhere
plt.rcParams['font.family'] = 'sans-serif'

# ORGANIZE - this can be deleted, it's only used in function plot_clustering_results below, which can be deleted
def create_clustering_function_values_list(clustering_data,locality):
    clustering_results=[[] for x in range(1,locality)]
    for clusters, value in clustering_data.items():
        if type(clustering_data[clusters])==list:
            for item in clustering_data[clusters]:
                clustering_results[item[0]-2].append(item[1:3])
    return clustering_results

# ORGANIZE - this can be deleted, as it's only used in visualise_clustering_function below, which can be deleted
def plot_clustering_results(clustering_results_list):
    for item in clustering_results_list:
        alpha_list = []
        clustering_function_values_list = []
        for element in item:
            alpha_list.append(element[0])
            clustering_function_values_list.append(element[1])

        plt.plot(alpha_list, clustering_function_values_list ,'ro')
        plt.show()
        #print(alpha_list)
        #print(clustering_function_values_list)

# ORGANIZE - this isn't used anywhere
def visualise_clustering_function(clustering_data,locality):
    plot_clustering_results( create_clustering_function_values_list(clustering_data,locality))

# ORGANIZE - the same function is copied into functions_benchmarks and is only used in code as an import from there
# the function can be deleted here
# function that plots value of benchmark vs values of alpha for a given set of test/traning data
def create_plots(plot_x_list, plot_y_list, separable_value, label):
    legend = []
    for i in range(len(plot_x_list)):

        # this if exludes plotting results for uncorelated noise model, they are added as a line below
        if len(plot_x_list[i]) > 1:
            plt.plot(plot_x_list[i], plot_y_list[i], marker='o', linestyle='none')
            title = f"Mitigation " + label
            legend.append(f'Locality {i + 1}')
            min_val = min(plot_y_list[i])
            if min_val < 10 ** (-2):
                min_x = plot_x_list[i][plot_y_list[i].index(min_val)]
                plt.annotate(str(np.round(min_val, 5)), xy=(min_x, min_val))

            plt.title(title)
            fname = title + ".png"
    plt.ylim(-0.001, 0.075)

    legend.append("Uncorrelated")

    plt.plot([plot_x_list[1][0], plot_x_list[1][-1]], [separable_value, separable_value])
    plt.annotate(str(np.round(separable_value, 5)), xy=(2.5, separable_value))
    plt.legend(legend)
    plt.savefig(fname)

# ORGANIZE - the same function is copied into functions_benchmarks and is only used in code as an import from there
# the function can be deleted here
#function that prepares data for plots
#input: clusters dictionary: dictionary with clusters 2) mitigation results: dictionary with clusters and mitigation reults
#output: three 2d lists with the first index encoding locality of cluster and the second one: 1) values of alpha 2) values of median for a cluster corresponding to agiven cluster 3) values of mean for a cluster corresponding to agiven cluster
#eg plot_median_list[2][0] encodes clustering for locality 2+1=3 corresponding to the alpha value plot_alpha_list[2][0]
def cerate_data_alpha_plot(clusters_dictionary, mitigation_data):
    parameters_dictionary = {}
    # loop over cluster assigments from clustering algorithm
    for cluster, clustering_params in clusters_dictionary.items():
        # this dictionary will store localities and corresponding alphas for a given cluster as well as median and mean for clustering
        locality_dic = {}
        max_locality = 0

        if clustering_params != None:
            # this is a loop over different configurations of localities and alpha for a given cluster structure
            for item in clustering_params:
                # data is rewritten as dictionary of a form locality : list of alphas
                if item[0] > max_locality:
                    max_locality = item[0]
                if item[0] in locality_dic.keys():
                    temp_list = locality_dic[item[0]]
                    temp_list.append(item[1])
                    locality_dic[item[0]] = temp_list

                else:

                    locality_dic[item[0]] = [item[1]]

        else:
            locality_dic[1] = [1]

        # to a given cluster value of meadian and mean of mitigation benchmark is added
        locality_dic['meadian'] = mitigation_data[cluster]['median']
        locality_dic['mean'] = mitigation_data[cluster]['mean']
        parameters_dictionary[cluster] = locality_dic
        # print(parameters_dictionary[cluster])

    plot_alpha_list = [[] for i in range(2, max_locality + 2)]
    plot_median_list = [[] for i in range(2, max_locality + 2)]
    plot_mean_list = [[] for i in range(2, max_locality + 2)]

    for clusters, parameters in parameters_dictionary.items():
        for i in range(1, max_locality + 1):
            if i in parameters.keys():
                for item in parameters[i]:
                    plot_alpha_list[i - 1].append(item)
                    plot_median_list[i - 1].append(parameters['meadian'])
                    plot_mean_list[i - 1].append(parameters['mean'])

    return plot_alpha_list, plot_median_list, plot_mean_list


# function calculates mean and median of mitigation benchmarks over a test/traning set
# input: benchmarks_results_mitigation - dictionary with mitigation results, traning_set - indicies of hamiltonians that form the traning set

#ORGANIZE - this function can be deleted; it is copied into functions_benchmarks and is only used as an import from there
def calculate_results_test_set(benchmark_results_mitigation, hamiltonian_set):
    # loop over differnt noise models (keys of benchmark_results_mitigation dictionary) and benchmark results for individual hamiltonians
    traning_median_mean = {}
    for noise_model, result in benchmark_results_mitigation.items():
        median, mean = calculate_median_mean(result['errors_list'], hamiltonian_set)
        traning_median_mean[noise_model] = {'median': median, 'mean': mean}
    return traning_median_mean

#ORGANIZE - this function can be deleted; it is copied into functions_benchmarks and is only used there
# helper function that calculates median and mean for a single noise model
def calculate_median_mean(errors_dictionary, hamiltonian_set):
    median_list = []
    mean = 0.0
    for index in hamiltonian_set:
        error = errors_dictionary[index]
        mean = mean + error
        median_list.append(error)
    mean = mean / len(hamiltonian_set)
    return np.median(median_list), mean


from typing import Optional, List, Dict, Union

from qrem.noise_characterization.base_classes.SDKHandlerBase import SDKHandlerBase

# TODO: check what do names other than ddot and qdot mean?
__valid_experiments_names__ = ['ddot',
                               'ddt',
                               'qdot',
                               'qdt',
                               'classical_states',
                               'separable_classical_states',
                               'separable_states',
                               'separable_quantum_states']



#(PP) This actually connnnects with the external SDKs: converts format to qiskit or pyqil and prepares circuits
class SeparableCircuitsCreator(SDKHandlerBase):
    """
    This is general class for creating quantum circuits implementing separable states on a quantum
    device. This is done based on provided list of circuits labels and chosen "interpreter" of those
    labels (see descriptions below).

    NOTE: please note that this class' is used only to create QuantumCircuit objects that can be passed
    to execution on some backend. It is assumed that specific labels describing circuits (see below)
    already have been generated.

    MAIN PARAMETERS:
    :param SDK_name:
    :param qubit_indices: one always needs to provide list of qubit indices on which experiments
                          should be performed (this does not need to be the whole device)

    :param descriptions_of_circuits: main argument to the class, this is list of symbolic circuits description

    Two formats are accepted:
    """
    # ORGANIZATION - in my opinion it is confucing that we have two arguments for qubit indices
    """
    1. List[List[int]] - integer indicates gate to be implemented on
                         corresponding qubit.
                         For example: circuit_labels = [[0,1,5],
                                                        [2,0,1]]

                         will implement circuits:
                         gate "0" on qubit_indices[0],
                         gate "1" on qubit_indices[1],
                         gate "5" on qubit_indices[2]

                         gate "2" on qubit_indices[0],
                         gate "0" on qubit_indices[1],
                         gate "1" on qubit_indices[2]

    2. Dict[str,int] - dictionary where each KEY is string of integers describing
                      a circuit (in manner analogous to described above),
                      and VALUE is the number of times that this circuit should be implemented

    Now the meaning of the 'gate i' depends on the type of experiment.
    For example, if one wishes to perform Diagonal Detector Overlapping Tomography (DDOT) [0.5],
    then there will be only two gates - idle and X gate, labeled by 0 and 1, respectively.

    In general, interpretation of symbols in descriptions_of_circuits parameter is done via something which
    we call here "circuits interpreter" (one of classes' properties).

    If one wishes to perform one of the available experiments (see __valid_experiments_names__ list),
    there is no need to specify such interpreter.

    One can add custom circuits interpreter using method "add_custom_circuits_interpreter".
    Alternatively, one can provide interpreter by passing suitable argument to method
    "__initialize_circuits_interpreter" (see description of the method).


    :param experiment_name: string describing type of experiment. It is used to initialize existing
                            circuits interpreters.

    possible values:

    1. 'DDOT' - implements Diagonal Detector Overlapping Tomography
    2. 'QDOT' - implements Quantum Detector Overlapping Tomography



    """

    def __init__(self,
                 SDK_name: str,
                 experiment_name: str,
                 qubit_indices: List[int],
                 descriptions_of_circuits: Union[Dict[str, int], List[List[int]]],
                 number_of_repetitions: Optional[int] = 1,
                 quantum_register_size: Optional[int] = None,
                 # MOcomm this variable does not seam to be used in relevant way - refactoring needed
                 classical_register_size: Optional[int] = None,
                 # MOcomm this variable does not seam to be used in relevant way - refactoring needed
                 add_barriers=True,
                 pyquil_compilation_method='parametric-native'):

        """
        :param SDK_name: see parent class' description
        :param experiment_name: see class' description
        #TODO FBM: check how important is that
        :param qubit_indices: see class' description
        :param descriptions_of_circuits: see class' description
        :param number_of_repetitions: each circuit in circuit_labels will be implemented
                                      this number of times
                                      NOTE: if there are doubles in the list, they will be counted as
                                        multiple circuits.
                                        For example: circuit_labels = [ [0,1,0], [0,1,0] ]
                                        with number_of_repetitions = 3
                                        will implement 6 circuits [0,1,0].

        #TODO FBM: check if this is used
        :param quantum_register_size: is used by method _create_circuit in parent class
        :param classical_register_size: is used by method _create_circuit in parent class
        """
        # ORGANIZEATION I checked this class. It does noy use quantum/ classical register size, its parent does
        # still this class is porely documented
        """

        :param add_barriers:
        """

        super().__init__(SDK_name=SDK_name,
                         qubit_indices=qubit_indices,
                         descriptions_of_circuits=descriptions_of_circuits,
                         experiment_name=experiment_name,
                         number_of_repetitions=number_of_repetitions,
                         quantum_register_size=quantum_register_size,
                         classical_register_size=classical_register_size,
                         add_barriers=add_barriers,
                         pyquil_compilation_method=pyquil_compilation_method)

        # TODO FBM, JT: maybe modify to include more than Pauli
        if self._experiment_name.lower() in ['ddot',
                                             'ddt',
                                             'classical_states',
                                             'separable_classical_states']:
            self._circuits_interpreter = self.__circuits_interpreter_DDOT
        elif self._experiment_name.lower() in ['qdot',
                                               'qdt',
                                               'separable_quantum_states']:
            self._circuits_interpreter = self.__circuits_interpreter_QDOT
        else:

            raise ValueError(
                f"Experiments of type: '{self._experiment_name}' is not supported yet.")
        # print(self._circuits_interpreter)
        self._experiment_name = experiment_name

    @property
    def circuits_interpreter(self):
        return self._circuits_interpreter

    @circuits_interpreter.setter
    def circuits_interpreter(self, circuits_interpreter):
        self._circuits_interpreter = circuits_interpreter

    @property
    def circuit_labels_dictionary(self) -> Dict[str, int]:
        return self._circuit_labels

    # TODO FBM: opisac to lepiej i dodac ze to konkretnie dla Pauliego
    def __circuits_interpreter_DDOT(self,
                                    circuit_label_list,
                                    circuit_object,
                                    qubit_indices):
        """
        This method creates DDOT circuits in format native to the given SDK (specified by class attribute SDK_name),
        based on the passed circuit_label_list and qubit_indices, specifying what gates should be applied to what qubits.
        """

        # TODO FBM: make sure qreg does not need to be passed
        # [X] This loop goes through all passed circuits and translates them to qiskit format, into variable circuit_object
        if self._SDK_name.upper() in ['QISKIT']:
            for qubit_index in range(len(circuit_label_list)):
                qubit_now = qubit_indices[qubit_index]
                label_now = circuit_label_list[qubit_index]

                if label_now in [0, '0']:
                    circuit_object.id(qubit_now)
                elif label_now in [1, '1']:
                    circuit_object.x(qubit_now)
                else:
                    raise ValueError('Wrong circuit label: ', circuit_label_list)

        # [X] This loop goes through all passed circuits and translates them to pyquil format, into variable circuit_object
        # This loop is almost identical to the one above, they could be merged into one and separation into SDKs
        # could be done inside the loop
        elif self._SDK_name.upper() in ['PYQUIL', 'PYQUIL-FOR-AZURE-QUANTUM']:
            if self.pyquil_compilation_method.lower() in ['high_level']:
                from pyquil.gates import I, X
                for qubit_index in range(len(circuit_label_list)):
                    qubit_now = qubit_indices[qubit_index]
                    label_now = circuit_label_list[qubit_index]

                    if int(label_now) == 0:
                        circuit_object += I(qubit_now)
                    elif int(label_now) == 1:
                        circuit_object += X(qubit_now)
                    else:
                        raise ValueError('Wrong circuit label: ', circuit_label_list)
            elif self.pyquil_compilation_method.lower() in ['parametric-native']:
                from pyquil.quilbase import Declare
                from pyquil.quilatom import MemoryReference
                from pyquil.gates import RX, RZ

                for qubit_index in qubit_indices:
                    circuit_object += Declare(name=f'rz-{0}_q-{qubit_index}',
                                              memory_type='REAL',
                                              memory_size=1)

                for qubit_index in qubit_indices:
                    circuit_object += RX(MemoryReference(f'rz-{0}_q-{qubit_index}', qubit_index))
        return circuit_object

    # TODO FBM: opisac to lepiej i dodac ze to konkretnie dla Pauliego
    def __circuits_interpreter_QDOT(self,
                                    circuit_label_list,
                                    circuit_object,
                                    qubit_indices):
        if self._SDK_name.upper() in ['QISKIT']:
            for qubit_index in range(len(circuit_label_list)):
                qubit_now = qubit_indices[qubit_index]
                label_now = circuit_label_list[qubit_index]

                # TODO FBM: make sure qreg does not need to be passed
                if int(label_now) == 0:
                    circuit_object.id(qubit_now)
                elif int(label_now) == 1:
                    circuit_object.x(qubit_now)
                elif int(label_now) == 2:
                    circuit_object.h(qubit_now)
                elif int(label_now) == 3:
                    circuit_object.x(qubit_now)
                    circuit_object.h(qubit_now)
                elif int(label_now) == 4:
                    circuit_object.h(qubit_now)
                    circuit_object.s(qubit_now)
                elif int(label_now) == 5:
                    circuit_object.x(qubit_now)
                    circuit_object.h(qubit_now)
                    circuit_object.s(qubit_now)
                else:
                    raise ValueError(
                        f"Wrong circuit label: '{label_now} from circuit: '{circuit_label_list}'")


        elif self._SDK_name.upper() in ['PYQUIL', 'PYQUIL-FOR-AZURE-QUANTUM']:

            if self.pyquil_compilation_method.lower() in ['high_level']:
                from pyquil.gates import I, X, H, S
                for qubit_index in range(len(circuit_label_list)):
                    qubit_now = qubit_indices[qubit_index]
                    label_now = circuit_label_list[qubit_index]

                    if int(label_now) == 0:
                        circuit_object += I(qubit_now)
                    elif int(label_now) == 1:
                        circuit_object += X(qubit_now)
                    elif int(label_now) == 2:
                        circuit_object += H(qubit_now)
                    elif int(label_now) == 3:
                        circuit_object += X(qubit_now)
                        circuit_object += H(qubit_now)
                    elif int(label_now) == 4:
                        circuit_object += H(qubit_now)
                        circuit_object += S(qubit_now)
                    elif int(label_now) == 5:
                        circuit_object += X(qubit_now)
                        circuit_object += H(qubit_now)
                        circuit_object += S(qubit_now)
                    else:
                        raise ValueError('Wrong circuit label: ', circuit_label_list)

            elif self.pyquil_compilation_method.lower() in ['parametric-native']:
                from pyquil.quilbase import Declare
                from pyquil.quilatom import MemoryReference
                from pyquil.gates import RX, RZ
        # ORGANIZATION: this part of code is commented - I suggest to removing it (MO)
        # ORGANIZATION: this part is supposed to give us full control of what gates are used, instead of relying on the
        # compiler - I suggest we discuss (JM)
        # for qubit_index in qubit_indices:
        #     for angle_name in [f'rz-{0}', f'rx-{0}', f'rz-{1}', f'rx-{1}', f'rz-{0}']:
        #         circuit_object += Declare(name=f'{angle_name}_q-{qubit_index}',
        #                                   memory_type='REAL',
        #                                   memory_size=1)
        #
        # for qubit_index in qubit_indices:
        #     for angle_name in [f'rz-{0}', f'rx-{0}', f'rz-{1}', f'rx-{1}', f'rz-{0}']:
        #         if angle_name[1]=='z':
        #             circuit_object += RZ(MemoryReference(f'{angle_name}_q-{qubit_index}',qubit_index))
        #         elif angle_name[1]=='x':
        #             circuit_object += RX(
        #                 MemoryReference(f'{angle_name}_q-{qubit_index}', qubit_index))
        #         else:
        #             raise ValueError(f"WRONG ANGLE NAME {angle_name}")

        return circuit_object

    # def __initialize_circuits_interpreter(self,circuits_interpreter):

    def get_circuits(self,
                     # TODO JT: break this function down so it gets single circuit
                     # circuit_object:Optional[] = None
                     add_measurements: Optional[bool] = True):  # MOcomm not sure what is the meaning of this variable
        # TODO JT: modify this to allow also for process tomography 
        # MOcomm - most likelly we will do it defferently
        """

        Returns quantum circuits as a list.

        Circuits are later identified by names for which we use the following convention:

        circuit_name = "experiment name" + "-" + "circuit label"+
        "no"+ "integer identifier for multiple implementations of the same circuit"

        for example the circuit can have name:
        "DDOT-010no3"

        which means that this experiment is Diagonal Detector Overlapping Tomography (DDOT),
        the circuit implements state "010" (i.e., gates iden, X, iden on qubits 0,1,2), and
        in the whole circuits sets this is the 4th (we start counting from 0) circuit that implements
        that particular state.

        :param add_measurements:
        :return:
        """

        qubit_indices = self._qubit_indices

        quantum_register_size, classical_register_size = self._quantum_register_size, \
            self._classical_register_size

        qubit_indices = sorted(qubit_indices)  # Sort to ensure, that results can easily be interpreted.

        if self._SDK_name.lower() in ['pyquil', 'pyquil-for-azure-quantum']:
            if self.pyquil_compilation_method.lower() in ['parametric-native',
                                                          'parametric',
                                                          'native-parametric']:
                from qrem.backends_support.pyquil import pyquil_utilities

                list_of_circuit_labels = self._circuits_labels_list

                memory_map = pyquil_utilities.create_memory_map_DOT(list_of_keys=list_of_circuit_labels,
                                                                    qubit_indices=qubit_indices)

                return memory_map

        # why to copy local variables to some extra variables
        descriptions_of_circuits = self._circuit_labels

        if quantum_register_size is None:
            quantum_register_size = max(qubit_indices) + 1

        if classical_register_size is None:
            classical_register_size = len(qubit_indices)

        all_circuits = []

        unique_circuit_labels = list(descriptions_of_circuits.keys())

        # outer loop is for copies of DDOT experiment
        # for repetition_index in range(number_of_repetitions):
        # inner loop goes through all circuits in QDT experiments and prepares them
        for circuit_label_string in unique_circuit_labels:
            # this loop goes over multiple instances of the same experiments (if they exist)
            circuit_label_list = list(circuit_label_string)

            for multiple_circuits_counter in range(descriptions_of_circuits[circuit_label_string]):

                # [X] create quantum circuit with nice name:
                circuit_name = self._experiment_name + "-" \
                               + circuit_label_string + 'no%s' % multiple_circuits_counter
                # [X] create "empty" circuit object for the given SDK:
                circuit_object, quantum_register, classical_register = self._create_circuit(
                    classical_register_size=classical_register_size,
                    quantum_register_size=quantum_register_size,
                    qubit_indices=qubit_indices,
                    circuit_name=circuit_name)
                # [X] add appropriate gates to the "empty" circuit:
                circuit_object = self._circuits_interpreter(circuit_label_list,
                                                            circuit_object,
                                                            qubit_indices)



                # TODO JM, FBM: add braket-qiskit and pyquil-azure implementation

                if self._SDK_name.upper() in ['QISKIT']:
                    if self._add_barriers:
                        circuit_object.barrier()
                    if add_measurements:
                        # TODO JM, FBM: add single abstract method for adding measurements?
                        circuit_object = self._add_measurements_qiskit(circuit_object=circuit_object,
                                                                       qreg=quantum_register,
                                                                       creg=classical_register,
                                                                       qubit_indices=qubit_indices)



                elif self._SDK_name.upper() in ['PYQUIL', 'PYQUIL-FOR-AZURE-QUANTUM']:
                    if add_measurements:
                        circuit_object = self._add_measurements_pyquil(quantum_program=circuit_object,
                                                                       qubit_indices=qubit_indices,
                                                                       classical_register=classical_register)

                all_circuits.append(circuit_object)

        return all_circuits



"""
@author: Jan Tuziemski, Filip Maciejewski, Joanna Majsak, Oskar Słowik, Marcin Kotowski, Katarzyna Kowalczyk-Murynka, Paweł Przewłocki, Piotr Podziemski, Michał Oszmaniec
@contact: filip.b.maciejewski@gmail.com
"""
from typing import List
import numpy as np

from qrem.functions_qrem import ancillary_functions as anf

from qrem.common.printer import qprint
from qrem.common import convert

def get_parallel_tomography_on_non_overlapping_subsets(number_of_circuits: int,
                                                       number_of_symbols: int,
                                                       non_overlapping_subsets: List[List[int]],
                                                       ) -> List[List[int]]:
    """
    :param number_of_circuits: should be power of number_of_symbols
    :param non_overlapping_subsets:
    :param number_of_symbols:
    :return:
    :rtype:
    """


    lengths_list = [len(x) for x in non_overlapping_subsets]
    number_of_qubits = sum(lengths_list)
    unique_lengths = np.unique(lengths_list)
    maximal_length = int(np.max(unique_lengths))


    if number_of_circuits % number_of_symbols != 0:
        number_of_circuits2 = -int(number_of_circuits // -maximal_length) * maximal_length
        qprint(
            f"Number of circuits should be power of {number_of_symbols}.\nChanging number of circuits from {number_of_circuits} to {number_of_circuits2}")
        number_of_circuits = number_of_circuits2

    if number_of_circuits<number_of_symbols**maximal_length:
        number_of_circuits2 = int(number_of_symbols**maximal_length)
        qprint(f"Number of circuits should be at least {number_of_symbols}^{maximal_length}={number_of_symbols**maximal_length} for subsets of locality at most {maximal_length}.\nChanging number of circuits from {number_of_circuits} to {number_of_circuits2}")
        number_of_circuits = number_of_circuits2

    local_registers = {length: convert.get_ditstrings_register(base=number_of_symbols,
                                                           number_of_dits=length)
                       for length in unique_lengths}

    enumerated_subsets = dict(enumerate(non_overlapping_subsets))

    circuits_subsets = []
    rng = np.random.default_rng()
    for subset in non_overlapping_subsets:
        local_register = local_registers[len(subset)]
        number_of_repetitions = number_of_circuits // len(local_register)
        all_states_local = local_register * number_of_repetitions
        rng.shuffle(all_states_local)
        circuits_subsets.append(all_states_local)

    circuits_tuples = list(zip(*circuits_subsets))
    circuits_list_integers = []
    for circ_tuple in circuits_tuples:
        circuit_integers = np.zeros((number_of_qubits),dtype=int)
        for subset_index in range(len(circ_tuple)):
            subset_now = enumerated_subsets[subset_index]
            local_circ_now = circ_tuple[subset_index]

            for qubit_index in range(len(subset_now)):
                circuit_integers[subset_now[qubit_index]] = local_circ_now[qubit_index]

        circuit_integers = list(circuit_integers)
        circuits_list_integers.append(circuit_integers)

    return circuits_list_integers

# number_of_qubits = 150
# number_of_symbols=5
# number_of_circuits = 100*number_of_symbols**2
#
# subsets = [[2*i,2*i+1] for i in range(number_of_qubits//2)]
# # print(subsets)
#
# circuits_list = get_parallel_tomography_on_non_overlapping_subsets(number_of_circuits=number_of_circuits,
#                                          number_of_symbols=number_of_symbols,
#                                          non_overlapping_subsets=subsets)
# # print(circuits_list)







import itertools
from typing import List, Dict, Optional, Callable, Tuple

import numpy as np
import math
import scipy.special as sp

from qrem.types import CircuitCollection
from qrem.common import constants
from qrem.common.experiment.tomoutils import algortithmic_circuits


def compute_number_of_circuits(experiment_type: str, 
                               number_of_qubits: int, 
                               subset_locality: Optional[int],
                               effect_distance: Optional[float] = constants.CIRC_EFFECT_DISTANCE,
                               probability_of_fail: Optional[float] =  constants.CIRC_PROBABILITY_OF_FAIL,
                               limited_circuit_randomness: Optional[bool] = False,
                               max_provider_circuit_count: Optional[int] = 4000 )-> dict:
    """Computes the optimal number of circuits for the experiment preparation stage

    Parameters:
        experiment_type (str): DDOT or QDOT (Diagonal/Quantum Detector Overlapping Tomography), defines the set of symbols used to encode the gates
        number_of_qubits (int): number of qubits in the experiment
        subset_locality (int): number of qubits on which the characterization will be performed, default value 0 means all qubits
        effect_distance (float): the tolerated distance between the estimated and actual effect operators
        probability_of_fail (float): probability of estimation error larger than effect_distance
        limited_circuit_randomness (bool): true if there is a limited number of circuits allowed by a provider
        maximum_provider_circuit_count (int):

    Returns:
        the number of circuits for which the probability of error larger  than effect_distance is bounded by probability_of_fail
    Raises:
        an error for a wrong experiment_type

    """

    try:
        number_of_symbols = constants.EXPERIMENT_TYPE_SYMBLOS[experiment_type.lower()]
    except:
        print("Error: wrong experiment type")

    if subset_locality==None:
        subset_locality = number_of_qubits
    n = number_of_qubits
    k = subset_locality
    s = number_of_symbols
    total_circuits_count = int(s**k*(np.log(math.factorial(n)/(math.factorial(k)*math.factorial(n-k)))+(k+1)*np.log(2)+np.log(1./probability_of_fail))/(2*effect_distance**2))
    random_circuits_count = min(max_provider_circuit_count,total_circuits_count)
    number_of_repetitions = max(1,math.ceil(total_circuits_count/(max_provider_circuit_count)))

    return {'number_of_repetitions': number_of_repetitions,
            'total_circuits_count': total_circuits_count,
            'random_circuits_count': random_circuits_count}

def create_random_circuits(experiment_type: str, number_of_qubits: int, number_of_circuits: int)-> List:
    """Creates a list of random mcircuits for the experiment preparation stage

    Parameters:
        experiment_type (str): DDOT or QDOT (Diagonal/Quantum Detector Overlapping Tomography), defines the set of symbols used to encode the gates
        number_of_qubits (int): number of qubits in the experiment
        number_of_circuits (int): maximal number of circuits
    Returns:
        a list of circuits, each given by a list of symbols encoding the gates acting on every qubit
    Raises:
        an error for a wrong experiment_type
    """

    try:
        number_of_symbols = constants.EXPERIMENT_TYPE_SYMBLOS[experiment_type.lower()]
    except:
        print("Error: wrong experiment type")
    
    circuits = np.random.randint(0, number_of_symbols,(number_of_circuits,number_of_qubits), dtype=constants.CIRCUIT_DATA_TYPE)


    return circuits

def check_completeness(experiment_type: str, circuit_list: List, locality = 2, symbols_list:Optional[List] = None,subsets_list = None):
    #TODO: write docstring (KKM)

    try:
        number_of_symbols = constants.EXPERIMENT_TYPE_SYMBLOS[experiment_type.lower()]
    except:
        print("Error: wrong experiment type")

    if symbols_list==None:
        symbols_list = list(range(number_of_symbols))
    number_of_qubits = len(circuit_list[0])
    if subsets_list == None:
        subsets = list(itertools.combinations(range(number_of_qubits),locality))
    else:
        subsets = subsets_list


    circuits = np.array(circuit_list)
    absent_symbols_dict = {}
    for t in subsets:
        locality = len(t)
        symbols = list(itertools.product(symbols_list, repeat=locality))
        indices = np.array(t)
        circuits_on_subset = list(map(tuple,circuits[:,indices]))
        absent = set(symbols).difference(circuits_on_subset)
        if absent != set():
            absent_symbols_dict[t] = list(absent)
    return absent_symbols_dict



#TODO: WORK IN PROGESS (PP)
def generate_circuits(number_of_qubits:int,
                      experiment_type:str = "ddot", 
                      k_locality:Optional[int] = 2,
                      add_random_circuits:Optional[bool] = True,
                      symbols:Optional[List] = None,
                      check_completness:Optional[bool] = True,
                      limited_circuit_randomness:Optional[bool] = False,
                      imposed_max_random_circuit_count:Optional[int] = 4000,
                      imposed_max_number_of_shots:Optional[int] = 1000):
    
    #[1] First we need to compute nubmer of circuits we should use.
    # Either we will be able to send all of them, or not; if not - we want to know how many we loose
    circuit_count_dict = compute_number_of_circuits(experiment_type =experiment_type, 
                                                    number_of_qubits = number_of_qubits, 
                                                    subset_locality=k_locality,
                                                    limited_circuit_randomness = limited_circuit_randomness,
                                                    max_provider_circuit_count = imposed_max_random_circuit_count)
    
    number_of_shots = circuit_count_dict['number_of_repetitions']
    total_circuit_count = circuit_count_dict['total_circuits_count']
    random_circuit_count = circuit_count_dict['random_circuits_count']
    if(random_circuit_count != total_circuit_count):
        print(f"WARNING: Possible count of random circuits ({random_circuit_count}) is lower than desired total circuit count ({total_circuit_count}).")
    #[2] Then we need to generate a hopefully complete set of circuits, even on this low amount
    # - can we estimate what will be a minimal complete set?
    #TODO.... finish based on generate_circuits(circuit_collection: CircuitCollection)
    algorithm_circuits = algortithmic_circuits.generate_combinatorial_circuits(experiment_type=experiment_type,number_of_qubits=number_of_qubits,subset_locality=k_locality,symbols=symbols)

    if not add_random_circuits:
        return algorithm_circuits, len(algorithm_circuits), total_circuit_count, number_of_shots


    if(check_completness):
        circuits = algorithm_circuits.copy()
        incomplete_dict = check_completeness(experiment_type=experiment_type,circuit_list=circuits,locality=k_locality,symbols_list=symbols)
        print(f"completeness: {incomplete_dict == {}}")
        while incomplete_dict != {}:
            if len(circuits)+len(list(incomplete_dict.values())[0])<=random_circuit_count:
                circuit_list_update = algortithmic_circuits.complete_on_subset(experiment_type, circuits,
                                                                list(incomplete_dict.keys())[0],
                                                                list(incomplete_dict.values())[0])

                dict_update = check_completeness(experiment_type, circuit_list_update, k_locality)
                circuits = circuit_list_update
                incomplete_dict = dict_update
            else:
                subsets = math.factorial(number_of_qubits)/(math.factorial(k_locality)*math.factorial(number_of_qubits-k_locality))
                print(f"WARNING: Set of circuits of size {len(circuits)} still incomplete on {100*len(incomplete_dict.keys())/subsets}% of subsets")
                #what to to if len(circuits)==rcc but still incomplete?
        #        Future development: avoid this by
        #        - completing on partitions
        #        - algorithmic completeness for k=3,4
                break
        circuit_count = random_circuit_count - len(circuits)
        if circuit_count>0:
            random_circuits = create_random_circuits(experiment_type=experiment_type,
                                          number_of_qubits=number_of_qubits,
                                          number_of_circuits=circuit_count)
            print(f"Adding {circuit_count} random circuits to a {len(circuits)}-element set")
            circuits = np.append(circuits, random_circuits, axis=0)

    else:
        circuit_count = random_circuit_count - len(algorithm_circuits)
        circuits = create_random_circuits(experiment_type=experiment_type,
                                   number_of_qubits=number_of_qubits,
                                   number_of_circuits=circuit_count)
        print(random_circuit_count, circuit_count)
        circuits = np.append(circuits, algorithm_circuits, axis=0)
    incomplete_dict = check_completeness(experiment_type=experiment_type, circuit_list=circuits, locality=k_locality)
    print(f"Set of {len(circuits)} circuits, completeness: {incomplete_dict=={}}")
    if len(circuits)>random_circuit_count:
        print(f"WARNING: cutting off {len(circuits)-random_circuit_count} last circuits")
    return circuits[0:random_circuit_count], random_circuit_count, total_circuit_count, number_of_shots



def batch_circuits(circuits_list:List, circuits_per_job:int=300):
    '''
    Function creating list of batches in abstract format. It takes the following inputs
        circuits list - list with tomographic circuits
        circuits_per_job - number of circuits in a batch 
    '''

    total_number_of_circuits = len(circuits_list)

    #computes total number of batches
    number_of_jobs = int(np.ceil(total_number_of_circuits / circuits_per_job))

    #initializes baches set
    batches = []
    #iteration over batches_index and creation of baches of size 
    for batch_index in range(number_of_jobs):
        circuits_now = circuits_list[batch_index * circuits_per_job : (batch_index + 1) * circuits_per_job] 
        batches.append(circuits_now)
    return batches
    # Because of how python treats lists/ arrays the last bach has automatically the right size (of the remaining circuits)

"""
    
def generate_circuits(circuit_collection: CircuitCollection):
    #TODO: 1. write docstring (KKM)
    #TODO: 2. reference to yaml provider configs
    collection_dict = circuit_collection.get_dict_format()
    experiment_type = collection_dict['experiment_type']
    number_of_qubits = len(collection_dict['qubit_indices'])
    locality = collection_dict['locality']

    circuit_count_dict = compute_number_of_circuits(experiment_type, number_of_qubits, locality)
    random_circuit_count = circuit_count_dict['random_circuits_count']
    total_circuit_count = circuit_count_dict['total_circuits_count']

    circuit_list = create_circuits(experiment_type=experiment_type,number_of_qubits=number_of_qubits,number_of_circuits=random_circuit_count)
    incomplete_dict = check_completeness(experiment_type=experiment_type,circuit_list=circuit_list,locality=locality)

    while incomplete_dict!={}:
        circuit_list_update = complete_on_subset(experiment_type, circuit_list, list(incomplete_dict.keys())[0], list(incomplete_dict.values())[0])
        dict_update = check_completeness(experiment_type, circuit_list_update, locality)
        circuit_list = circuit_list_update
        incomplete_dict = dict_update

    random_circuit_count = len(circuit_list)

    return circuit_list, random_circuit_count, total_circuit_count
"""

if __name__ == "__main__":
    exp_types = ['coh']
    number_of_qubits = [128]
    locality = [2]
    for e in exp_types:
        for n in number_of_qubits:
            for k in locality:
                print(f"Experiment type: {e}, {n} qubits, locality {k}")
                print(compute_number_of_circuits(e,n,k))
                c = generate_circuits(number_of_qubits=n, experiment_type=e, k_locality=k,add_random_circuits=True,symbols=[4,5])










import numpy as np
from qrem.ctmp.mitigtools.observables_utils import convert, random_element, sample, evaluate, delta
from qrem.functions_qrem import povmtools
from qrem.types import CTMPModelData
from typing import Dict, Tuple, List, Union
from qrem.common import probability




def _mitigate(model: CTMPModelData,
              output_counts: Dict[str, int],
              input_obs,
              T: int,
              mode: str,
              proj: bool=False) -> np.array:
    """ Mitigate either an observable or marginal, defined by mode. 

    Parameters
    ----------
    Input_obs: Union[List,Tuple]
         Input_obs = observable given as list of coefficients  if mode=obs or a tuple of qubits if mode=marginal
    """
    # todo: check input correctness
    # check statistical correctness
    # M = sum(output_counts.values())
    c1 = np.exp(2*model.gamma)
    if mode == "obs":
        xi = np.zeros(T)
        # convert observable description into an array
        obs = convert(input_obs, model.n)
    if mode == "marginal":
        xi = np.zeros((T, 2**len(input_obs)))
    # Algorithm 1 from the paper
    for t in range(T):
        alpha = np.random.poisson(model.gamma)
        s = random_element(output_counts)
        #i = np.random.randint(0, M)
        x = sample(model, alpha, s) #string of 0s and 1s representing output sample
        if mode == "obs":
            x = np.array([int(v) for v in x]) # maybe change this later
            xi[t] = (-1)**alpha * evaluate(obs, x)
        if mode == "marginal":
            xi[t] = (-1)**alpha * delta(x, input_obs)
    result = xi.mean(axis=0) * c1
    if mode == "marginal" and proj:
        result = result / result.sum() # normalize to sum 1
        if not probability.is_valid_probability_vector(result):
            result = povmtools.find_closest_prob_vector_l2(result).flatten()
    return result



# -*- coding: utf-8 -*-
"""
Created on Wed Jun  1 10:39:15 2016

@author: Peter Wittek
"""

from __future__ import division, print_function
import numpy as np
from scipy.sparse import lil_matrix
from .sdpa_utils import convert_row_to_sdpa_index


def solve_with_cvxpy(sdp, solverparameters=None):
    """Helper function to convert the SDP problem to CVXPY
    and call the solver, and parse the output.

    :param sdp: The SDP relaxation to be solved.
    :type sdp: :class:`ncpol2sdpa.sdp`.
    """
    problem = convert_to_cvxpy(sdp)
    if solverparameters is not None and 'solver' in solverparameters:
        solver = solverparameters.pop('solver')
        v = problem.solve(solver=solver, verbose=(sdp.verbose > 0), **solverparameters)
    else:
        v = problem.solve(verbose=(sdp.verbose > 0))
    if v is None:
        status = "infeasible"
        x_mat, y_mat = [], []
    elif v == float("inf") or v == -float("inf"):
        status = "unbounded"
        x_mat, y_mat = [], []
    else:
        status = "optimal"
        x_pre = sdp.F[:, 1:sdp.n_vars+1].dot(problem.variables()[0].value)
        x_pre = x_pre.reshape(x_pre.shape + (1,))
        x_pre += sdp.F[:, 0]
        row_offsets = [0]
        cumulative_sum = 0
        for block_size in sdp.block_struct:
            cumulative_sum += block_size ** 2
            row_offsets.append(cumulative_sum)
        x_mat = []
        for bi, bs in enumerate(sdp.block_struct):
            x = x_pre[row_offsets[bi]:row_offsets[bi+1]].reshape((bs, bs))
            x += x.T - np.diag(np.array(x.diagonal())[0])
            x_mat.append(x)
        y_mat = [constraint.dual_value for constraint in problem.constraints]
    return v+sdp.constant_term, v+sdp.constant_term, x_mat, y_mat, status


def convert_to_cvxpy(sdp):
    """Convert an SDP relaxation to a CVXPY problem.

    :param sdp: The SDP relaxation to convert.
    :type sdp: :class:`ncpol2sdpa.sdp`.

    :returns: :class:`cvxpy.Problem`.
    """
    from cvxpy import Minimize, Problem, Variable
    row_offsets = [0]
    cumulative_sum = 0
    for block_size in sdp.block_struct:
        cumulative_sum += block_size ** 2
        row_offsets.append(cumulative_sum)
    x = Variable(sdp.n_vars)
    # The moment matrices are the first blocks of identical size
    constraints = []
    for idx, bs in enumerate(sdp.block_struct):
        nonzero_set = set()
        F = [lil_matrix((bs, bs)) for _ in range(sdp.n_vars+1)]
        for ri, row in enumerate(sdp.F.rows[row_offsets[idx]:
                                            row_offsets[idx+1]],
                                 row_offsets[idx]):
            block_index, i, j = convert_row_to_sdpa_index(sdp.block_struct,
                                                          row_offsets, ri)
            for col_index, k in enumerate(row):
                value = sdp.F.data[ri][col_index]
                F[k][i, j] = value
                F[k][j, i] = value
                nonzero_set.add(k)
        if bs > 1:
            sum_ = sum(F[k]*x[k-1] for k in nonzero_set if k > 0)
            if not isinstance(sum_, (int, float)):
                if F[0].getnnz() > 0:
                    sum_ += F[0]
                constraints.append(sum_ >> 0)
        else:
            sum_ = sum(F[k][0, 0]*x[k-1] for k in nonzero_set if k > 0)
            if not isinstance(sum_, (int, float)):
                sum_ += F[0][0, 0]
                constraints.append(sum_ >= 0)
    obj = sum(ci*xi for ci, xi in zip(sdp.obj_facvar, x) if ci != 0)
    problem = Problem(Minimize(obj), constraints)
    return problem



"""
Classes for visualisation of detector tomography results.
"""
from pathlib import Path
import copy
import math

from manim import Scene, Graph, Arrow, NumberLine, Rectangle, VGroup, Tex, Text
from manim import tempconfig
from manim import BLACK, DEGREES, LEFT
import matplotlib.pyplot as plt
import numpy as np

from qrem.visualisation.plotting_constants import *

#==============================================================
# Support classes
#==============================================================
class DrawOnLayout(Scene):
    """
    Class inheriting from manim class Scene (for details on this convention see manim documentation).
    Images are rendered when this class' instance is created in the method 'draw_on_layout'
    of the class ResultsForPlotting.

    Parameters
    ----------
        plotting_data (dict):
            dictionary with data about the device and processed experimental results, necessary for manim graph
            (for details see manim graph documentation), with values:
                vertices (list[int]) : list of qubits in device, from 0 to #qubits - 1.
                edges (list[tuple[int, int]]) : list of edges (i, j) to be plotted, where i and j are in vertices.
                labels (bool|dict): dictionary whose keys are the vertices, and whose values are the corresponding
                                    vertex labels (rendered via, e.g., Text or manim Tex). Alternatively can be bool -
                                    if True, vertices are labelled with the integers from the list 'vertices'.
                vertex_config (dict): dictionary with keyword arguments which apply to manim mobject used for vertices.
                edge_config (dict): dictionary with keyword arguments which apply to manim mobject used for edges.
                layout (dict[list]): dictionary where keys are vertices and values are list of coordinates [x, y, z].
                (Optional) max_for_legend (float): maximal value to be displayed on the legend bar;
                                                   if not in keys, no legend is displayed.
    """

    def __init__(self, plotting_data):
        super().__init__()
        self.plotting_data = plotting_data

    def construct(self):
        layout_graph = Graph(self.plotting_data['vertices'],
                             self.plotting_data['edges'],
                             labels=self.plotting_data['labels'],
                             label_fill_color=WHITE,
                             vertex_config=self.plotting_data['vertex_config'],
                             edge_type=Arrow,
                             edge_config=self.plotting_data['edge_config'],
                             layout=self.plotting_data['layout']).scale(self.plotting_data['scale'])

        layout_graph.center()
        self.add(layout_graph)

        if 'max_for_legend' in self.plotting_data.keys():
            legend = NumberLine(x_range=[0, self.plotting_data['max_for_legend'], 0.05],
                                length=legend_bar_height,
                                include_tip=False,
                                include_numbers=True,
                                include_ticks=True,
                                rotation=90 * DEGREES,
                                color=BLACK,
                                stroke_opacity=0.5,
                                label_direction=LEFT,
                                font_size=20)
            legend.align_on_border([1, 0, 0], 0.5 + legend_bar_width / 2 + 0.05)
            legend.numbers.set_color(BLACK)
            legend_bar = Rectangle(
                height=legend_bar_height,
                width=legend_bar_width,
                fill_color=list(colors),
                fill_opacity=1,
                stroke_opacity=0
            )
            legend_bar.align_on_border([1, 0, 0], 0.5)
            self.add(legend_bar)
            self.add(legend)

class DrawClusters(Scene):
    """
    Class inheriting from manim class Scene (for details on this convention see manim documentation).
    Images are rendered when this class' instance is created in the method 'draw_clusters'
    of the class ResultsForPlotting.

    Parameters
    ----------
        clusters_list (list(tuples[ints])):
            list of tuples, each containing indices of qubits assigned to one cluster
    """

    def __init__(self, clusters_list):
        super().__init__()
        self.clusters_list = clusters_list

    def construct(self):
        graphs = []
        for cluster in self.clusters_list:
            vertices = list(cluster)
            edges = []
            for idx in range(len(vertices)):
                jdx = idx + 1
                while jdx < len(vertices):
                    edges.append((vertices[idx], vertices[jdx]))
                    jdx += 1
            graphs.append(Graph(vertices, edges, layout="circular", labels=True, label_fill_color=WHITE,
                                vertex_config={"fill_color": to_hex(color_map(0.5)), "radius": radius_size},
                                edge_config={"stroke_color": to_hex(color_map(0.5))}).scale(0.7))
        self.add(VGroup(*graphs).arrange())


#==============================================================
# Main Class
#==============================================================device_data_divctionary
class ResultsForPlotting:
    """
    Class whose attributes contain experimental tomography data necessary for plotting. Its method prepare the data in
    a format native to manim and create an instance of its class, which results in an image being rendered and saved.
    TODO: add remark about indexing over all vs used qubits
    """

    def __init__(self,
                 device_data_dictionary,
                 experiment_data_dictionary):
        """
        Parameters
        ----------
        device_data_dictionary (dict): device data, can be provided from file device_constants or manually. Keys:
            no_qubits (int) : number of qubits in device.
            edges_in_layout (list[tuple[int, int]]) : list of edges (i, j) present in device layout,
                                                      where 0 <= i, j < no_qubits.
            layout (dict[list]): dictionary where keys are vertices and values are list of coordinates [x, y, z].
            vertex_labels (list[int]): list of length no_qubits with qubits labels (e.g. in Rigetti devices 9th qubit
                                       has label 10) in same order as layout dictionary
            scale (float): fraction by which the whole layout graph will be scaled to fit the default manim window
        experiment_data_dictionary (dict) : experiment data. Keys:
            used_qubits (list[int]) : indices of qubits which were used in tomography (not discarded due to noise).
            correlations (ndarray(used_qubits, used_qubits)) : array of correlation coefficients between qubits,
                                                               where [i][j] indicates the influence of qubit j on i.
        Attributes
        ----------
            correlations_threshold (float) : threshold of correlations coefficient above which the corresponding edges
                                             should be plotted as coloured arrows.
            edges_with_correlations_above_threshold (list[tuple[int, int]])
            clusters_list (list[tuple[int, ..., int]]) : list of clusters into which qubits were divided during
                                                          characterisation


        """
        config.tex_dir = Path('Tex')  # manim configuration - path where tex files created for plot will be saved

        self.no_qubits = device_data_dictionary['no_qubits']
        self.edges_in_layout = device_data_dictionary['edges_in_layout']
        self.layout = device_data_dictionary['layout']
        self.scale = device_data_dictionary['scale']
        self.all_qubits = list(range(self.no_qubits))
        self.all_edges = [val for sublist in [[(i, j) for i in range(self.no_qubits)]
                                              for j in range(self.no_qubits)]
                          for val in sublist]
        self.labels = {}
        for qubit_idx in self.all_qubits:
            self.labels[qubit_idx] = Tex(str(device_data_dictionary['vertex_labels'][qubit_idx]))
        self.used_qubits = experiment_data_dictionary['used_qubits']
        # transpose correlations array, so that correlations[i][j] indicates the influence of qubit i on j:
        self.correlations = np.transpose(experiment_data_dictionary['correlations'])
        self.correlations_threshold = None
        self.edges_above_threshold = []
        self.max_correlation_coefficient = np.amax(self.correlations)
        self.clusters_list = None

    # def __identify_max_correlations(self):
    #     self.edges_directed_by_max_correlations = {}
    #     for qubit0 in range(len(self.used_qubits)):
    #         for qubit1 in range(qubit0 + 1, len(self.used_qubits)):
    #             # map indices of used qubits to indices in device layout
    #             mapped_qubit_pair = (self.used_qubits[qubit0], self.used_qubits[qubit1])
    #             if self.correlations[qubit0][qubit1] >= self.correlations[qubit1][qubit0]:
    #                 self.edges_directed_by_max_correlations[mapped_qubit_pair] = self.correlations[qubit0][qubit1]
    #             else:
    #                 self.edges_directed_by_max_correlations[tuple(reversed(mapped_qubit_pair))] = \
    #                     self.correlations[qubit1][qubit0]

    def __identify_correlations_above_threshold(self, threshold):
        self.edges_above_threshold = []
        self.correlations_threshold = threshold
        for qubit0 in range(len(self.used_qubits)):
            for qubit1 in range(len(self.used_qubits)):
                if self.correlations[qubit0][qubit1] >= threshold:
                    self.edges_above_threshold.append((qubit0, qubit1))

    def __prepare_basic_edge_config(self):
        basic_edge_config = {}
        for idx, e in enumerate(self.edges_in_layout):
            if e[0] in self.used_qubits and e[1] in self.used_qubits:
                edge_color = medium_gray
            else:
                edge_color = lightest_gray
            basic_edge_config[e] = {"stroke_color": edge_color, "stroke_width": stroke_width_thin, "buff": buff_val,
                                    "max_tip_length_to_length_ratio": 0.0, }
        return basic_edge_config

    def __prepare_basic_vertex_config(self):
        basic_vertex_config = {}
        for idx, v in enumerate(self.all_qubits):
            basic_vertex_config[v] = {"color": lightest_gray, "radius": radius_size, "z_index": -1}
            if v in self.used_qubits:
                basic_vertex_config[v] = {"color": medium_gray, "radius": radius_size}
        return basic_vertex_config

    def __prepare_edge_config_for_correlations_above_threshold(self, correlations_threshold):
        self.__identify_correlations_above_threshold(correlations_threshold)
        edge_config = self.__prepare_basic_edge_config()
        for edge in self.edges_above_threshold:
            qubit0 = edge[0]
            qubit1 = edge[1]
            mapped_qubit0 = self.used_qubits[qubit0]
            mapped_qubit1 = self.used_qubits[qubit1]
            stroke_color = to_hex(color_map(self.correlations[qubit0][qubit1]
                                            / self.max_correlation_coefficient))
            # map indices of used qubits to indices in device layout
            mapped_qubit_pair = (mapped_qubit0, mapped_qubit1)
            edge_config[mapped_qubit_pair] = {"stroke_color": stroke_color, "stroke_width": stroke_width_thick,
                                              "max_stroke_width_to_length_ratio": max_stroke_width_to_length_ratio,
                                              "max_tip_length_to_length_ratio": arrow_tip_size / (
                                                      0.01 + math.dist(self.layout[mapped_qubit0],
                                                                       self.layout[mapped_qubit1])) ** 1.4}
        return edge_config

    def __prepare_vertex_config_for_correlations_above_threshold(self):
        vertices_with_correlations_above_threshold = []

        for edge in self.edges_above_threshold:
            vertex_affecting = self.used_qubits[edge[0]]
            vertex_affected = self.used_qubits[edge[1]]
            if vertex_affected not in vertices_with_correlations_above_threshold:
                vertices_with_correlations_above_threshold.append(vertex_affected)
            if vertex_affecting not in vertices_with_correlations_above_threshold:
                vertices_with_correlations_above_threshold.append(vertex_affecting)

        vertex_config = self.__prepare_basic_vertex_config()
        for idx, v in enumerate(vertices_with_correlations_above_threshold):
            vertex_config[v] = {"color": darkest_gray, "radius": radius_size, "z_index": 0}
        return vertex_config

    def __prepare_data_for_draw_basic_layout(self):
        plotting_data = {'vertices': self.all_qubits,
                         'edges': self.edges_in_layout,
                         'labels': self.labels,
                         'layout': self.layout,
                         'scale': self.scale,
                         'vertex_config': self.__prepare_basic_vertex_config(),
                         'edge_config': self.__prepare_basic_edge_config()}
        return plotting_data

    def __prepare_data_for_draw_correlations_on_layout(self, threshold):
        labels = copy.deepcopy(self.labels)
        for qubit_idx in range(self.no_qubits):
            if qubit_idx not in self.used_qubits:
                # the following makes the label of unused qubit a transparent dot, using manim class Text :
                labels[qubit_idx] = Text('.', fill_opacity=0.0)

        edge_config = self.__prepare_edge_config_for_correlations_above_threshold(threshold)

        # create a list with edges where the edges from basic layout are below those above threshold:
        mapped_edges_above_threshold = [(self.used_qubits[unmapped_edge[0]], self.used_qubits[unmapped_edge[1]])
                                        for unmapped_edge in self.edges_above_threshold]
        edges = copy.deepcopy(self.edges_in_layout)
        for edge in mapped_edges_above_threshold:
            if edge in edges:
                edges.remove(edge)
            if tuple(reversed(edge)) in edges:
                edges.remove(tuple(reversed(edge)))
        edges = edges + mapped_edges_above_threshold

        plotting_data = {'vertices': self.all_qubits,
                         'edges': edges,
                         'labels': labels,
                         'layout': self.layout,
                         'scale': self.scale,
                         'edge_config': edge_config,
                         'vertex_config': self.__prepare_vertex_config_for_correlations_above_threshold(),
                         'max_for_legend': self.max_correlation_coefficient}
        return plotting_data

    def set_clusters_list(self, clusters_list):
        self.clusters_list = clusters_list

    def draw_on_layout(self, what_to_draw='basic_layout', file_name=None, quality='high_quality', preview=True,
                       correlations_threshold=0.01, verbosity='CRITICAL'):
        """
        Method called by user to plot the experimental data on the device's layout.

        Parameters
        ----------
            what_to_draw (str) : specifies what data should be plotted. Options:
                                    'basic_layout' : qubits with labels and edges in layout are plotted.
                                    'correlations_above_threshold' : on basic layout also edges with correlations
                                                                     coefficient above a given threshold are plotted and
                                                                     colored, with a color legend.
            quality (str) : quantifies the quality of rendered image. For details see manim docs.
                            [fourk_quality|production_quality|high_quality|medium_quality|low_quality|example_quality].
            preview (bool) : specifies whether the rendered image should be displayed in a user's default file viewer.
                             Regardless of this parameter's value, the image is saved.
            file_name (str): name of the output image file. By default, it's set to 'what_to_draw'
            correlations_threshold (float) : parameter for what_to_draw set to 'correlations_above_threshold',
                                             in range [0, 1].
            verbosity (str) : Verbosity of CLI output. [DEBUG|INFO|WARNING|ERROR|CRITICAL]
        """
        if what_to_draw == 'basic_layout':
            plotting_data = self.__prepare_data_for_draw_basic_layout()
        elif what_to_draw == 'correlations_above_threshold':
            plotting_data = self.__prepare_data_for_draw_correlations_on_layout(correlations_threshold)
        if file_name is None:
            file_name = what_to_draw
        # run manim class:
        with tempconfig({"quality": quality, "preview": preview, "output_file": file_name, "verbosity": verbosity}):
            scene = DrawOnLayout(plotting_data)
            scene.render(preview=preview)

    def draw_histogram(self, file_name='histogram', directory='', plot_title='Correlations histogram',
                       plot_label='Distance name'):
        """
            Method called by user to plot correlation coefficients on a histogram.
            Saves the plot with 'file_name' in 'directory'.

        """
        plt.rc('text', usetex=True)
        plt.rcParams.update({"font.family": "Computer Modern Roman"})
        correlations_for_hist = copy.deepcopy(self.correlations)
        # delete diagonal elements (corresponding to qubit correlation with itself) and flatten:
        correlations_for_hist = np.delete(correlations_for_hist,
                                          [k*len(self.correlations) + k for k in range(len(self.correlations))], None)
        fig, ax = plt.subplots(figsize=(8, 5))
        n, bins, patches = ax.hist(correlations_for_hist, bins=10, density=False, label=plot_label,
                                   color=color_map(0.1), linewidth=1.6)
        ax.legend(loc='upper right')
        ax.set_title(plot_title)
        ax.set_xlabel('Correlation coefficient')
        ax.set_ylabel('Number of qubit pairs')
        plt.savefig(directory + file_name + '.png', bbox_inches='tight', dpi=300)

    def draw_heatmap(self, file_name='heatmap', directory='', plot_title='Correlation coefficients'):
        """
            Method called by user to plot correlation coefficients on a heatmap.
            Saves the plot with 'file_name' in 'directory'.

        """
        plt.rc('text', usetex=True)
        plt.rcParams.update({"font.family": "Computer Modern Roman"})
        fig, ax = plt.subplots()
        im = ax.imshow(self.correlations, cmap=color_map)
        cbar = ax.figure.colorbar(im, ax=ax)
        ax.spines[:].set_visible(False)
        cbar.outline.set_visible(False)
        ax.set_title(plot_title)
        ax.set_xticks(np.arange(0, self.correlations.shape[1], 1))
        ax.set_yticks(np.arange(0, self.correlations.shape[0], 1))
        ax.set_xticklabels(self.used_qubits)
        ax.set_yticklabels(self.used_qubits)
        ax.tick_params(which="minor", bottom=False, left=False)
        plt.savefig(directory + file_name + '.png', bbox_inches='tight', dpi=300)

    def draw_clusters(self, file_name='clusters', quality='high_quality', preview=True, verbosity='CRITICAL'):
        """
            Method called by user to plot the clusters in abstract form.
        Parameters
        ----------
            quality (str) : quantifies the quality of rendered image. For details see manim docs.
                            [fourk_quality|production_quality|high_quality|medium_quality|low_quality|example_quality].
            preview (bool) : specifies whether the rendered image should be displayed in a user's default file viewer.
                             Regardless of this parameter's value, the image is saved.
            file_name (str): name of the output image file. By default, it's set to 'what_to_draw'
            verbosity (str) : Verbosity of CLI output. [DEBUG|INFO|WARNING|ERROR|CRITICAL]
        """
        if self.clusters_list is not None:
            with tempconfig({"quality": quality, "preview": preview, "output_file": file_name, "verbosity": verbosity}):
                scene = DrawClusters(self.clusters_list)
                scene.render(preview=preview)
        else:
            raise ValueError('First set value of clusters_list attribute')



"""
@authors: Filip Maciejewski, Oskar Słowik, Tomek Rybotycki
@contact: filip.b.maciejewski@gmail.com
"""

import copy
import math
import random
import qutip
import numpy as np
import scipy as sc
import re
from typing import Dict
import matplotlib.pyplot as plt

from qrem.functions_qrem.quantum_ancillary_functions import get_energy_from_bitstring_diagonal



def l1_norm(v):
    sum = 0
    for x in v:
        sum += abs(x)
    return sum


def delta(v, w):
    val = 0
    if v == w:
        val = 1
    return val

#(PP) possible overlap with bitstrings
def generate_binary_strings(n, l):
    if n == 0:
        return l
    else:
        if len(l) == 0:
            return generate_binary_strings(n - 1, ["0", "1"])
        else:
            return generate_binary_strings(n - 1, [i + "0" for i in l] + [i + "1" for i in l])

#(PP) dobuled in other files
#(PP) part of a model of Bravi noise
#(PP) should go to qrem.noise_modelling
def return_1local_gen(type, r):
    if type == '0':
        gen = [[-r, 0], [r, 0]]
    elif type == '1':
        gen = [[0, r], [0, -r]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


#(PP) dobuled in other files
def return_2local_gen(type, r):
    if type == '00':
        gen = [[-r, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [r, 0, 0, 0]]
    elif type == '01':
        gen = [[0, 0, 0, 0], [0, -r, 0, 0], [0, r, 0, 0], [0, 0, 0, 0]]
    elif type == '10':
        gen = [[0, 0, 0, 0], [0, 0, r, 0], [0, 0, -r, 0], [0, 0, 0, 0]]
    elif type == '11':
        gen = [[0, 0, 0, r], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, -r]]
    else:
        raise ValueError('Wrong generator type!')
    return np.array(gen)


'''
s - a list of numpy vectors representing measurement results
O - a numpy vector with values O(x)
S - a list of numpy arrays representing operatrs S_alpha
c - a list of weights of each S_alpha
delta - delta parameter
'''


def bravyi_algorithm_1_naive(s, O, S, c, delta):
    c_l1_norm = l1_norm(c)
    # print(c_l1_norm)
    M = len(s)  # number of experiments
    # print(M)
    d = O.shape[0]  # dimension
    # print(d)
    T = math.ceil(4 * c_l1_norm ** 2 / (delta ** 2))
    # print(T)
    xi = np.zeros(T)
    q_alpha = [abs(x) / c_l1_norm for x in c]
    # print(q_alpha)
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        alpha = np.random.choice(np.arange(0, len(c)), p=q_alpha)
        # print(alpha)
        q_x = S[alpha].dot(s[i])
        # print(q_x)
        x = np.random.choice(np.arange(0, d), p=q_x)
        # print(x)
        xi[t] = np.sign(c[alpha]) * O[x]
        # print(xi[t])
    return c_l1_norm * np.sum(xi) / T


'''
O=np.array([0.2,0.3,-1,1])
S=[np.array([[0,0,0,0],[1,0,0,0],[0,0,0,1],[0,1,1,0]]),np.array([[1,0,0,0],[0,1,0,0],[0,0,1,0],[0,0,0,1]])]
s=[np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,1,0]),np.array([0,0,0,1])]
c=[0.95,0.05]
delta=0.1
print(bravyi_algorithm_1(O,S,s,c,delta))
'''


# p(y_0...y_k | x_0 ... x_n), k<=n
def get_cond_prob(x, y, G_dict, lambda_par, printing=False):
    if printing: print('look for ' + str(x) + '->' + str(y))
    val = 0
    rank = len(y)
    support = set(range(rank))
    if printing: print('supp: ' + str(support))
    val += delta(y, x[:rank])
    if val == 1:
        if printing: print('DIAGONAL')
    for op_support_tuple in G_dict:
        operator = G_dict[op_support_tuple]
        if printing: print(operator)
        op_support_ordered = list(op_support_tuple)
        if printing: print(op_support_ordered)

        op_support_cap_support = set(op_support_ordered).intersection(set(support))
        if not op_support_cap_support == set():
            support_cond = True
            if printing: print('oscs: ' + str(op_support_cap_support))
            support_minus_op_support = set(support).difference(set(op_support_ordered))
            for qubit in support_minus_op_support:
                if y[qubit] != x[qubit]:
                    support_cond = False
                    if printing: print('sc failed for qubit' + str(qubit))
                    break
            if (support_cond):
                if printing: print('sc OK')
                op_support_minus_support = set(op_support_ordered).difference(set(support))
                if printing: print('osms: ' + str(op_support_minus_support))
                y_supp_vals = dict()

                for qubit in op_support_cap_support:
                    y_supp_vals[qubit] = y[qubit]

                no_of_y_summed = len(op_support_minus_support)

                if (no_of_y_summed > 0):
                    y_summed_bitstrings = generate_binary_strings(no_of_y_summed, [])
                    for y_summed_bitstr in y_summed_bitstrings:
                        i = 0
                        if printing: print('checking bitstr' + str(y_summed_bitstr))
                        for qubit in op_support_minus_support:
                            y_supp_vals[qubit] = y_summed_bitstr[i]
                            i += 1
                        y_supp_vals_ordered = "".join("%s" % y_supp_vals[qubit] for qubit in op_support_ordered)
                        row = int(y_supp_vals_ordered, 2)
                        x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in op_support_ordered)
                        col = int(x_supp_vals_ordered, 2)
                        if printing: print('added val ' + str(operator[row][col] / lambda_par))
                        val += operator[row][col] / lambda_par
                else:
                    y_supp_vals_ordered = "".join("%s" % y[qubit] for qubit in op_support_ordered)
                    row = int(y_supp_vals_ordered, 2)
                    x_supp_vals_ordered = "".join("%s" % x[qubit] for qubit in op_support_ordered)
                    col = int(x_supp_vals_ordered, 2)
                    if printing: print('no added bitstr')
                    if printing: print('added val ' + str(operator[row][col] / lambda_par))
                    val += operator[row][col] / lambda_par

    return val


'''
G_dict = dict()
r01_00 = 0.1
r01_11 = 0.05
r01_01 = 0.05

r12_00 = 0.1
r12_11 = 0.05
r12_01 = 0.05

# test_x=[[0,0,1],[0,0,0],[1,1,1]]
G_dict['q0q1'] = [[-r01_00, 0, 0, r01_11], [0, -r01_01, 0, 0], [0, r01_01, 0, 0], [r01_00, 0, 0, -r01_11]]
G_dict['q0q2'] = [[-r12_00, 0, 0, r12_11], [0, -r12_01, 0, 0], [0, r12_01, 0, 0], [r12_00, 0, 0, -r12_11]]

N = 3

x_set = generate_binary_strings(N, [])
y_set = generate_binary_strings(N - 1, [])
lambda_par = 0.4
'''
'''
process_mat=dict()
for x in x_set:
    for y in y_set:
        process_mat[(x,y)]=get_cond_prob(x, y, G_dict, lambda_par)
        print(x+'->'+y+' '+str(process_mat[(x,y)]))


ser = pd.Series(list(process_mat.values()),
                  index=pd.MultiIndex.from_tuples(process_mat.keys()))
df = ser.unstack().fillna(0).transpose()
sns.heatmap(df)
plt.savefig("output2.png")
'''


def sample_from_B_pow_n(n, x, G_dict, lambda_par, printing=False):
    sample = x
    for step in range(n):
        if printing: print('pre ' + sample)
        sample = sample_from_B(sample, G_dict, lambda_par)
        if printing: print('post ' + sample)
    return sample


def sample_from_B(x, G_dict, lambda_par, printing=False):
    y = ''
    no_of_bits = len(x)
    for bit_no in range(no_of_bits):
        y_0 = y + '0'
        y_1 = y + '1'
        if printing: print('testing ' + y_0 + ' and ' + y_1)
        cp_0 = get_cond_prob(x, y_0, G_dict, lambda_par)
        cp_1 = get_cond_prob(x, y_1, G_dict, lambda_par)
        tot = cp_0 + cp_1
        p_0 = cp_0 / tot
        if printing: print('p_0 ' + str(p_0))
        val = random.uniform(0, 1)
        if val < p_0:
            y = y + '0'
            if printing: print('extending ' + y)
        else:
            y = y + '1'
            if printing: print('extending ' + y)
    return y


def binary(num, length):
    return format(num, '0{}b'.format(length))


def sample_from_B_naive(x, B):
    dim = len(B[0])
    n = int(math.log(dim, 2))
    B = np.array(B)
    # print(x)
    # supp_str=''.join(x)
    # col=int(supp_str,2)
    col = int(x, 2)
    # print(col)
    prob_vector = B[:, col]
    row = np.random.choice(range(dim), p=prob_vector)
    # print('q'+str(n))
    sample = binary(row, n)
    # print('s'+str(sample))
    return sample


def sample_from_B_pow_n_naive(n, x, B, printing=False):
    sample = x
    for step in range(n):
        if printing: print('pre ' + sample)
        sample = sample_from_B_naive(sample, B)
        if printing: print('post ' + sample)
    return str(sample)


# sample_from_B_pow_n(10, x_set[0], G_dict, lambda_par)

'''
O = {'q0q1': 10, 'q1': -1, 'q2' : -1, 'q3' : -1 }
s = ['0000'] * 50
s2 = ['1111'] * 50
s.extend(s2)
delta_par = 0.1
'''

'''
s - a dict with experimental results
O - a dict with keys - pauli string supports x and values - coeffs
delta - delta parameter
'''
#TODO: rewrite it to s- dict, not a list !!!
def bravyi_algorithm_1_cond_prob(s, O, G_dict, delta_par, lambda_par):
    c_l1_norm = math.exp(2 * lambda_par)
    M = len(s)  # number of experiments
    # print('M '+str(M))

    T = math.ceil(4 * (c_l1_norm ** 2) / (delta_par ** 2))
    print('T ' + str(T))
    sum_xi = 0
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        n = np.random.poisson(lambda_par)
        # print('n '+str(n))
        x = sample_from_B_pow_n(n, s[i], G_dict, lambda_par)
        print('x ' + x)

        if (n % 2) == 0:
            sign = 1.0
        else:
            sign = -1.0
        sum_xi += sign * get_energy_from_bitstring_diagonal(x, O)
        print(get_energy_from_bitstring_diagonal(x, O))
    return c_l1_norm * sum_xi / T

#TODO: finish this with Filip's sampling and s as a dict with counts, not a list
def bravyi_algorithm_1(s, O, G_dict, delta_par, lambda_par):
    c_l1_norm = math.exp(2 * lambda_par)
    M = len(s)  # number of experiments
    # print('M '+str(M))

    T = math.ceil(4 * (c_l1_norm ** 2) / (delta_par ** 2))
    print('T ' + str(T))
    sum_xi = 0
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        n = np.random.poisson(lambda_par)
        # print('n '+str(n))
        x = sample_from_B_pow_n(n, s[i], G_dict, lambda_par)
        print('x ' + x)

        if (n % 2) == 0:
            sign = 1.0
        else:
            sign = -1.0
        sum_xi += sign * get_energy_from_bitstring_diagonal(x, O)
        print(get_energy_from_bitstring_diagonal(x, O))
    return c_l1_norm * sum_xi / T


# print(bravyi_algorithm_1(s, O, G_dict, delta_par, lambda_par))

# y -bitstring
# N - no of samples
# G_dict - dict with local noise generators
# lambda_par - lambda parameter
def sample_from_bravyi_noise_model(y, N, G_dict, lambda_par):
    out = []
    for t in range(N):
        n = np.random.poisson(lambda_par)
        # print('n ' + str(n))
        out.append(sample_from_B_pow_n(n, y, G_dict, lambda_par))
    return out


def bravyi_noise_model_naive_distribution(y, G_dict):
    N = len(y)
    G = convert_G_dict_to_G(G_dict, N)
    # print(G)
    # print(np.sum(G, axis=0))
    A = sc.linalg.expm(G)
    # print(A)
    # print(np.sum(A, axis=0))
    y_index = int(y, 2)
    # print(y_index)
    pd_x = A[:, y_index]
    return pd_x


def construct_model_from_rates(r_dict):
    local_G_dictionary = dict()


    for op_support_tuple in r_dict:
        no_of_qubits = int(len(op_support_tuple) )
        dim = int(math.pow(2, no_of_qubits))
        gen_total = np.zeros((dim, dim))
        for op_type in r_dict[op_support_tuple]:
            r = r_dict[op_support_tuple][op_type]
            if len(op_type) == 1:
                gen = return_1local_gen(op_type, r)
            elif len(op_type) == 2:
                gen = return_2local_gen(op_type, r)
            else:
                raise ValueError('Wrong generator dimension in r_dict!')
            gen_total = gen_total + gen
        local_G_dictionary[op_support_tuple] = gen_total
    return local_G_dictionary


# r_dict = {'q0q1': {'00': 10, '11': 9, '01': 8}, 'q0q2': {'00': 7, '11': 6, '01': 5}}


# print(sample_from_bravyi_noise_model(s, 10, G_dict, lambda_par))


# def generate_bravyi_G_dict(rates,N):

def embed_1q_operator(N, gate, index):
    if index == 0:
        embed_operator = np.kron(gate, np.eye(int(2 ** (N - 1))))
        return embed_operator
    else:
        first_eye = np.eye(2 ** (index))
        second_eye = np.eye(2 ** (N - index - 1))

        embeded_operator = np.kron(np.kron(first_eye, gate), second_eye)

        return embeded_operator


def embed_operator(N, gate, indices=[0, 1]):
    N_small = int(np.log2(gate.shape[0]))

    if N_small == 1:
        return embed_1q_operator(N, gate, indices[0])

    qutip_object = qutip.Qobj(gate, dims=[[2, 2] for i in range(N_small)])

    return qutip.qip.expand_operator(qutip_object, N, indices).full()


# def extend_matrix_element(gen_supp_indices, entry, G, N, supp1='', supp2=''):
#    if len(supp2) == N:
#        row
#        # add element
#        return

#    if len(supp1) == N:
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1, supp2 + '0')
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1, supp2 + '1')
#    else:
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1 + '0', supp2)
#        extend_matrix_element(gen_supp_indices, entry, G, N, supp1 + '1', supp2)

def generate_random_r_dict(qubits_indices, bound):
    r_dict = dict()

    for qubit in qubits_indices:
        supp_1_local = (qubit,)
        r_dict_sub_1 = dict()
        for op_type_1 in ['0', '1']:
            r_dict_sub_1[op_type_1] = bound * np.random.random()
        r_dict[supp_1_local] = r_dict_sub_1

    for ind_1,qubit_1 in enumerate(qubits_indices):
        for qubit_2 in qubits_indices[ind_1+1:]:
            supp_2_local = (qubit_1, qubit_2)
            r_dict_sub_2 = dict()
            for op_type_2 in ['00', '01', '10', '11']:
                r_dict_sub_2[op_type_2] = bound * (np.random.random()+0.01)
            r_dict[supp_2_local] = r_dict_sub_2
    return r_dict

# q1 q2 ...
def convert_G_dict_to_G(G_dict, N):
    dim = int(math.pow(2, N))
    G = np.zeros((dim, dim))
    for gen_supp in G_dict:
        gen_supp_indices = list(gen_supp)
        operator = np.array(G_dict[gen_supp])
        # print(operator)
        operator_emb = embed_operator(N, operator, indices=gen_supp_indices)
        # print(np.shape(operator_emb))
        G = G + operator_emb
    return G

def convert_local_G_dict_to_local_G(G_dict, supp):
    N = len(supp)
    dim = int(math.pow(2, N))
    G = np.zeros((dim, dim))
    for gen_supp in G_dict:
        gen_supp_indices = [supp.index(pos) for pos in gen_supp]
        operator = np.array(G_dict[gen_supp])
        # print(operator)
        operator_emb = embed_operator(N, operator, indices=gen_supp_indices)
        # print(np.shape(operator_emb))
        G = G + operator_emb
    return G

def generate_random_G_dict(qubits_indices, bound):

    r_dict = generate_random_r_dict(qubits_indices, bound)
    G_dict = construct_model_from_rates(r_dict)

    # TODO OS: this can be done inside above function
    lambda_b = get_r_sum_from_r_dict(r_dict)
    return G_dict, lambda_b

def generate_random_local_G_dict(clusters, bound):

    local_G_dictionaries={}
    r_dict= {}

    local_G_matrices = {}

    for cluster in clusters:
        qubits_indices = list(cluster)
        #print(qubits_indices)
        r_dict_clust = generate_random_r_dict(qubits_indices, bound)
        #print(r_dict_clust)
        r_dict.update(r_dict_clust)
        G_dict_clust = construct_model_from_rates(r_dict_clust)
        #print(G_dict_clust)
        local_G_dictionaries[cluster]=G_dict_clust
        local_G_matrix = convert_local_G_dict_to_local_G(G_dict_clust, qubits_indices)
        local_G_matrices[cluster] = local_G_matrix


    return local_G_dictionaries,local_G_matrices, r_dict


'''
clusters=[(0,1,2),(3,4),(5,6,7,8),(9,),(10,)]
bound = 0.05
local_G_dicts,local_G_mat, r_dict=generate_random_local_G_dict(clusters, bound)



for G_dict in local_G_dicts:
    print("New G_dict")
    print(G_dict)
print(r_dict)

for G_mat in local_G_mat:
    print(G_mat)
    print("New G_mat")
    print(local_G_mat[G_mat])
'''

def get_r_sum_from_r_dict(r_dict):
    r_sum = 0
    for op_support_ordered_str in r_dict:
        for op_type in r_dict[op_support_ordered_str]:
            r = r_dict[op_support_ordered_str][op_type]
            r_sum = r_sum + r
    return r_sum


def test_bravyi_sampling(y, bound, no_of_samples):
    no_of_qubits = len(y)
    dim = int(math.pow(2, no_of_qubits))
    pd_x_approx = np.zeros(dim)
    G_dict, lambda_par = generate_random_G_dict([i for i in range(no_of_qubits)], bound)
    pd_x = bravyi_noise_model_naive_distribution(y, G_dict)
    samples = sample_from_bravyi_noise_model(y, no_of_samples, G_dict, lambda_par)
    for sample in samples:
        index = int(sample, 2)
        pd_x_approx[index] += 1
    pd_x_approx = pd_x_approx / len(samples)

    pointwise_error = np.absolute(pd_x_approx - pd_x)
    TVD = np.max(pointwise_error)
    print(f'TVD: {TVD:.5f}')
    # print(samples)
    # print(pd_x)
    # print(pd_x_approx)

    ind = np.arange(dim)
    # print(ind)
    width = 0.35
    plt.bar(ind, pd_x_approx, width, label='sampling_alg')
    plt.bar(ind + width, pd_x, width, label='true')

    plt.ylabel('p')
    plt.title('Bravyi sampling pd for y=' + y + ', samples= ' + str(no_of_samples) + f', TVD= {TVD:.5f}')

    xt = [s[::-1] for s in generate_binary_strings(no_of_qubits, [])]

    plt.xticks(ind + width / 2, xt)

    plt.legend(loc='best')
    plt.show()


# test_bravyi_sampling('0101',0.05, 10000)

'''
N = 3
r_d = generate_random_r_dict(N, 2)
print(r_d)
r_sum = get_r_sum_from_r_dict(r_d)
G_dict = construct_model_from_rates(r_d)
G = convert_G_dict_to_G(G_dict, N)
print(r_sum)
print(G)
'''


# print(convert_G_dict_to_G(G_dict, 3))
# G_d=construct_model_from_rates(r_dict)
# print(convert_G_dict_to_G(G_d, 3))


def get_CTMP_rates_from_results(results_dictionary_ddot: Dict[str, Dict[str, int]],
                                number_of_qubits: int):
    single_qubits = list(range(number_of_qubits))
    pairs = [(i, j) for i in range(number_of_qubits) for j in range(i + 1, number_of_qubits)]

    local_noise_matrices_CTMP = {pair: np.zeros((4, 4), dtype=float) for pair in pairs}

    for pair in pairs:
        pair_complement = list(set(single_qubits).difference(set(pair)))

        for global_input_state, results_dictionary in results_dictionary_ddot.items():
            marginal_input_state = ''.join([global_input_state[x] for x in pair])
            input_state_complement = [global_input_state[x] for x in pair_complement]

            for global_output_state, ticks in results_dictionary.items():
                marginal_output_state = ''.join([global_output_state[x] for x in pair])
                output_state_complement = [global_output_state[x] for x in pair_complement]

                if output_state_complement == input_state_complement:
                    # check if this is their convention!
                    local_noise_matrices_CTMP[pair][int(marginal_output_state, 2),
                                                    int(marginal_input_state, 2)] += ticks

    # normalize to stochastic matrices
    for pair in pairs:
        for k in range(4):
            if sum(local_noise_matrices_CTMP[pair][:, k]) != 0:
                local_noise_matrices_CTMP[pair][:, k] /= sum(local_noise_matrices_CTMP[pair][:, k])

    # Get G matrices
    # TODO: potential bug with logarithm branch
    # print(local_noise_matrices_CTMP[pair])
    G_matrices = {pair: sc.linalg.logm(local_noise_matrices_CTMP[pair]) for pair in pairs}

    # ancillary function
    def _chop_negatives(M):
        (m, n) = M.shape

        chopped_M = copy.deepcopy(M)
        for i in range(m):
            for j in range(n):
                if i != j and M[i, j] < 0:
                    chopped_M[i, j] = 0

        return chopped_M

    # Get G' matrices
    G_prime_matrices = {pair: _chop_negatives(G_matrices[pair]) for pair in pairs}

    rates_dictionary_2q = {f"q{pair[0]}q{pair[1]}": {'00': 0,
                                                     '01': 0,
                                                     '10': 0,
                                                     '11': 0} for pair in pairs}

    for pair in pairs:
        G_prime_matrix_now = G_prime_matrices[pair]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['00'] = G_prime_matrix_now[3, 0]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['01'] = G_prime_matrix_now[2, 1]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['10'] = G_prime_matrix_now[1, 2]
        rates_dictionary_2q[f"q{pair[0]}q{pair[1]}"]['11'] = G_prime_matrix_now[0, 3]

    rates_dictionary_1q = {f"q{q}": {'0': 0, '1': 0} for q in single_qubits}
    for qj in single_qubits:
        r0, r1 = 0, 0

        for q_other in list(set(single_qubits).difference(set([qj]))):
            G_prime_matrix_now = G_prime_matrices[tuple(sorted([qj, q_other]))]

            r0 += G_prime_matrix_now[2, 0] + G_prime_matrix_now[3, 1] \
                # +G_prime_matrix_now[1,0]+G_prime_matrix_now[3,2]

            r1 += G_prime_matrix_now[0, 2] + G_prime_matrix_now[1, 3] \
                # +G_prime_matrix_now[0, 1] + G_prime_matrix_now[2, 3]

        r0 /= 2 * (number_of_qubits - 1)
        r1 /= 2 * (number_of_qubits - 1)

        rates_dictionary_1q[f"q{qj}"]['0'] = r0
        rates_dictionary_1q[f"q{qj}"]['1'] = r1

    rates_dictionary = {**rates_dictionary_1q,
                        **rates_dictionary_2q}
    return rates_dictionary


def check_CTMP_consistency(no_of_qubits, bound, samples_per_input):
    r_dict = generate_random_r_dict(no_of_qubits, bound)
    G_dict = construct_model_from_rates(r_dict)

    lambda_par = get_r_sum_from_r_dict(r_dict)

    output = dict()

    for input in generate_binary_strings(no_of_qubits, []):
        ticks = dict()
        input_results = sample_from_bravyi_noise_model(input, samples_per_input, G_dict, lambda_par)

        for result in input_results:
            if result in ticks.keys():
                ticks[result] = ticks[result] + 1
            else:
                ticks[result] = 1
        output[input] = ticks

    r_dict_reconstr = get_CTMP_rates_from_results(output, no_of_qubits)
    return r_dict, r_dict_reconstr


def merge_local_to_global(local_dict, n):
    # print(local_dict)
    merged_list = np.empty(n, dtype='str')
    for supp in local_dict:
        val = local_dict[supp]
        for local_index, q in enumerate(supp):
            merged_list[int(q)] = val[local_index]
            # print(val[local_index])
    # print(merged_list)
    merged = ''.join(merged_list)
    return merged


def trim_str_to_supp_tuple(y, supp):
    y_trimmed = ''
    for q in supp:
        y_trimmed += y[q]
    return y_trimmed


# r_dict, r_dict_reconstr = check_CTMP_consistency(3, 0.05, 10000)
# print(r_dict)
# print(r_dict_reconstr)

# y -bitstring
# N - no of samples
# local_B_dict - dict with keys : local supports and keys : local models as tuples (gamma, B)
def sample_from_local_bravyi_noise_model(y, N, local_B_dict):
    no_of_qubits = len(y)
    out = []
    for t in range(N):
        sample = {}
        for supp in local_B_dict:
            gamma = (local_B_dict[supp])[0]
            B = (local_B_dict[supp])[1]
            n = np.random.poisson(gamma)
            # print('n ' + str(n))
            y_trimmed = trim_str_to_supp_tuple(y, supp)
            sample[supp] = sample_from_B_pow_n_naive(n, y_trimmed, B)
            # print(sample[supp])
        sample_merged = merge_local_to_global(sample, no_of_qubits)
        out.append(sample_merged)
    return out


#local_B_dict = {(0, 1): (10, [[0.9, 0, 0, 0.3], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0.1, 0, 0, 0.7]]),
#                (2, 3): (12, [[0, 0.5, 0.5, 0], [0.2, 0, 0, 0.4], [0.8, 0, 0, 0.6], [0, 0.5, 0.5, 0]])}
#y = '0011'
#N = 50
#print(sample_from_local_bravyi_noise_model(y, N, local_B_dict))


# supp=(0,2,3)
# print(trim_str_to_supp_tuple(y,supp))

def get_global_gamma_from_local_B_dict(local_B_dict):
    global_gamma = 0
    for cluster in local_B_dict:
        global_gamma += (local_B_dict[cluster])[0]
    return global_gamma


def bravyi_algorithm_1_local(s, O, local_B_dict, delta_par):
    no_of_qubits = len(s[0])
    gamma_par = get_global_gamma_from_local_B_dict(local_B_dict)
    c_l1_norm = math.exp(2 * gamma_par)
    M = len(s)  # number of experiments
    # print('M '+str(M))

    T = math.ceil(4 * (c_l1_norm ** 2) / (delta_par ** 2))
    print('T ' + str(T))
    sum_xi = 0
    for t in range(T):
        i = random.randrange(0, M)
        # print("i"+str(i))
        # n = np.random.poisson(lambda_par)
        # print('n '+str(n))
        sample = {}
        n_sum = 0
        for supp in local_B_dict:
            gamma = (local_B_dict[supp])[0]
            B = (local_B_dict[supp])[1]
            n = np.random.poisson(gamma)
            n_sum += n
            # print('n ' + str(n))
            y_trimmed = trim_str_to_supp_tuple(s[i], supp)
            sample[supp] = sample_from_B_pow_n_naive(n, y_trimmed, B)
            # print(sample[supp])
        x = merge_local_to_global(sample, no_of_qubits)

        #print('y ' + s[i])
        #print('x ' + x)

        if (n_sum % 2) == 0:
            sign = 1.0
        else:
            sign = -1.0
        sum_xi += sign * get_energy_from_bitstring_diagonal(x, O)
        #print(get_energy_from_bitstring_diagonal(x, O))
    return c_l1_norm * sum_xi / T

'''
local_B_dict = {(0, 1): (10, [[0.9, 0, 0, 0.3], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0.1, 0, 0, 0.7]]),
                (2, 3): (12, [[0, 0.5, 0.5, 0], [0.2, 0, 0, 0.4], [0.8, 0, 0, 0.6], [0, 0.5, 0.5, 0]])}
O = {'q0q1': 10, 'q1': -1, 'q2' : -1, 'q3' : -1 }
s = ['0000'] * 50
s2 = ['1111'] * 50
s.extend(s2)
delta_par = 0.1
print(bravyi_algorithm_1_local(s, O, local_B_dict, delta_par))
'''